import math
from typing import List

import pennylane as qml

from .types import Hamiltonian, Observable, ParitySequence, Tensor
from .utils import all_bin_sequences


def parity_all_hamiltonian(num_qubits: int, weights: Tensor) -> Hamiltonian:
    """
    Hamiltonian corresponding to the parity of all combinations of Pauli Z operators.

    :param num_qubits: Number of qubits.
    :type num_qubits: int
    :param weights: Observable weights of shape (2^num_qubits,)
    :type weights: Tensor
    :return: Observable corresponding to the parity of Pauli Z operators
    :rtype: Observable
    :raises ValueError: If shape of weights is not as specified above.
    """

    shapeW = weights.shape
    if len(shapeW) != 1:
        raise ValueError(f"W (shape={shapeW}) must be a 1-dim tensor")

    W_qubits = math.log2(len(weights))
    if not W_qubits.is_integer():
        raise ValueError(f"Length of W ({W_qubits}) not a power of 2")

    if int(W_qubits) != num_qubits:
        raise ValueError(f"num_qubits and W (shape={shapeW}) do not match")

    obs = parities_all_observables(num_qubits)
    H = qml.Hamiltonian(weights, obs)

    return H


def parities_all_observables(n: int) -> List[Observable]:
    """
    Generates a list of observables corresponding to the parity of all
    possible binary combinations of n qubits.

    :param n: The number of qubits.
    :type n: int
    :return: A list of observables corresponding to the parity of all
        possible binary combinations of n qubits plus Idenity.
    :rtype: List[Observable]
    """

    seq = all_bin_sequences(n)
    return sequence2parity_observable(seq)


def sequence2parity_observable(parity_sequence: ParitySequence) -> List[Observable]:
    """
    Generates a list of observables corresponding to the parity of
    the indeces defined by the sequence.

    :param parity_sequence: The sequence of Z parities.
    :type parity_sequence: ParitySequence
    :return: A list of observables corresponding to the parity of all
        possible binary combinations of n qubits plus Idenity.
    :rtype: List[Observable]
    """

    ops = []
    for par in parity_sequence:
        if par:
            tmp = qml.PauliZ(par[0])
            if len(par) > 1:
                for i in par[1:]:
                    tmp = tmp @ qml.PauliZ(i)

            ops.append(tmp)

    ops.append(qml.Identity(0))

    return ops



from typing import Tuple

import torch

from .types import Tensor


class HatBasis:
    """
    Linear 1D hat basis for the range [a, b] and a specified number of nodes,
    including a and b.

    :param a: Left end point.
    :type a: float
    :param b: Right end point.
    :type b: float
    :param num_nodes: Number of nodes.
    :type num_nodes: int
    """

    def __init__(self, a: float, b: float, num_nodes: int) -> None:
        self.a = a
        self.b = b
        self.num_nodes = num_nodes
        self.num_segments = num_nodes - 1
        self.segment_length = (b - a) / self.num_segments

    def position(self, x: Tensor) -> Tensor:
        """
        Find the index of the grid point left of x.

        :param x: A tensor containing the values for which the position indexes are to be found.
        :type x: Tensor
        :returns: A tensors of position indexes.
        The position indices are -1 for values left of `a`, and -2 for values right of `b`.
        :rtype: Tensor
        """

        left_of_a = x < self.a
        right_of_b = x > self.b

        within_range = torch.logical_not(torch.logical_or(left_of_a, right_of_b))
        position = torch.zeros_like(x)
        position[within_range] = ((x[within_range] - self.a) / self.segment_length).floor()

        position[left_of_a] = -1
        position[right_of_b] = -2

        return position

    def grid_points(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        """
        Finds the grid points surrounding given values in the discretized space.

        :param x: A tensor containing the values for which the surrounding
        grid points are to be found.
        :type x: Tensor
        :returns: A tuple of two tensors. The first tensor contains the left boundary points
        of the segments, the second tensor contains the right boundary points of the segments.
        :rtype: Tuple[Tensor, Tensor]
        """

        left_point = torch.zeros_like(x)
        right_point = torch.zeros_like(x)

        left_of_a = x < self.a
        left_point[left_of_a] = self.a - self.segment_length
        right_point[left_of_a] = self.a

        right_of_b = x > self.b
        left_point[right_of_b] = self.b
        right_point[right_of_b] = self.b + self.segment_length

        within_range = torch.logical_not(torch.logical_or(left_of_a, right_of_b))

        position = self.position(x)

        left_point[within_range] = self.a + position[within_range] * self.segment_length
        right_point[within_range] = left_point[within_range] + self.segment_length

        return left_point, right_point

    def nonz_vals(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        """
        Output the 2 nonzero values for any given x.
        Finds the grid points surrounding given values in the discretized space.

        :param x: Input variables.
        :type x: Tensor
        :returns: A tuple of 2 nonzero values.
        :rtype: Tuple[Tensor, Tensor]
        """

        left, right = self.grid_points(x)
        first = (right - x) / self.segment_length
        second = (x - left) / self.segment_length
        return first, second

    def eval_basis_vector(self, x: Tensor) -> Tensor:
        """
        Evaluate basis vector at points x.

        :param x: Input variables.
        :type x: Tensor
        :returns: Vector of all basis functions evaluated at x.
        :rtype: Tuple[Tensor, Tensor]
        """

        nodes = torch.linspace(-1, 1, self.num_nodes, dtype=x.dtype)
        values = torch.zeros(x.shape[0], self.num_nodes, dtype=x.dtype)
        for i in range(1, self.num_nodes):
            mask = (nodes[i - 1] <= x) & (x <= nodes[i])
            values[mask, i - 1] = (nodes[i] - x[mask]) / self.segment_length
            values[mask, i] = (x[mask] - nodes[i - 1]) / self.segment_length

        return values.squeeze(0)



from abc import ABC, abstractmethod
from itertools import product
from typing import Generic, Optional, Set, Tuple, TypeVar

import numpy as np
import torch
from scipy.stats import qmc

from .types import (
    Array,
    CDevice,
    DataLoader,
    DataOut,
    Device,
    DType,
    Model,
    ParameterList,
    Tensor,
    TensorDataset,
)

D = TypeVar("D")
L = TypeVar("L")


class DataGenTorch(ABC, Generic[D, L]):
    """
    Abstract base class for generating data in PyTorch.

    :param batch_size: Size of batch for data loader. Defaults to None (batch mode).
    :type batch_size: int
    :param shuffle: Shuffle data in loader. Defaults to False.
    :type shuffle: bool
    :param seed: The seed used to initialize the random number generator.
        Defaults to None.
    :type seed: int
    :param device: The device where to run the computations. Defaults to
        torch.device("cpu").
    :type device: Device
    """

    def __init__(
        self,
        batch_size: Optional[int] = None,
        shuffle: bool = False,
        seed: Optional[int] = None,
        device: Device = torch.device("cpu"),
    ) -> None:
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.seed = seed
        self.device = device

    @abstractmethod
    def gen_data(self, *args, **kwargs) -> D:
        """
        Generate the data.

        :param args: Variable length argument list.
        :param kwargs: Arbitrary keyword arguments.
        :returns: The generated data.
        :rtype: D
        """
        pass

    @abstractmethod
    def data_to_loader(self, *args, **kwargs) -> L:
        """
        Convert data from gen_data to pytorch data loader.

        :param args: Variable length argument list.
        :param kwargs: Arbitrary keyword arguments.
        :returns: Data loader.
        :rtype: L
        """
        pass


class PriorTorch(ABC, Generic[D]):
    """
    Abstract base class for generating priors in PyTorch.

    :param sizex: Dimension of feature space.
    :type sizex: int
    :param seed: The seed used to initialize the random number
        generator. Defaults to None.
    :type seed: int, optional
    :param device: The device where to run the computations. Defaults to
        torch.device("cpu").
    :type device: Device
    """

    def __init__(
        self,
        sizex: int,
        seed: Optional[int] = None,
        device: Device = torch.device("cpu"),
    ) -> None:
        self.sizex = sizex
        self.seed = seed
        self.device = device

    @abstractmethod
    def gen_data(self, *args, **kwargs) -> D:
        """
        Generate the data.

        :param args: Variable length argument list.
        :param kwargs: Arbitrary keyword arguments.
        :returns: The generated data.
        :rtype: D
        """
        pass


class DataGenCapacity(DataGenTorch[DataOut, DataLoader]):
    """
    Generates data for memory capacity estimation.

    :param sizex: The size of the input data.
    :type sizex: int
    :param num_samples: The number of output samples to generate.
        Defaults to 10.
    :type num_samples: int, optional
    :param scale: The re-scaling factor for uniform random numbers in [0, 1].
        Defaults to 2.0.
    :type scale: float, optional
    :param shift: The shift value for uniform random numbers in [0, 1].
        Defaults to -1.0.
    :type shift: float, optional
    :param kwargs: Keyword arguments passed to the base class.
    """

    def __init__(
        self,
        sizex: int,
        num_samples: int = 10,
        scale: float = 2.0,
        shift: float = -1.0,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.sizex = sizex
        self.num_samples = num_samples
        self.scale = scale
        self.shift = shift

    def gen_data(self, N: int) -> DataOut:
        """
        Generate the data.

        :param N: The number of inputs for the model.
        :type N: int
        :return: A dictionary containing the generated data.
        :rtype: DataOut
        """

        X, Y = gen_dataset_capacity(
            N=N,
            sizex=self.sizex,
            num_samples=self.num_samples,
            seed=self.seed,
            scale=self.scale,
            shift=self.shift,
            device=self.device,
        )

        data = {"X": X, "Y": Y}

        return data

    def data_to_loader(self, data: DataOut, s: int) -> DataLoader:
        """
        Convert data to pytorch loader.

        :param data: Output of gen_data.
        :type data: DataOut
        :param s: Current label sample.
        :type s: int
        :rtype: DataLoader
        :raises ValueError: For invalid data or index s.
        """
        self._check_data(data)

        Ny = data["Y"].size(0)
        if s < 0 or s >= Ny:
            raise ValueError(f"Index i ({s}) must be between 0 and Ny ({Ny}).")

        X = data["X"]
        Y = data["Y"][s]
        Nx = X.size(0)
        dataset = TensorDataset(X, Y)
        batch_size = Nx
        if self.batch_size is not None:
            batch_size = self.batch_size

        loader = DataLoader(dataset, batch_size=batch_size, shuffle=self.shuffle)

        return loader

    def _check_data(self, data: DataOut):
        if "X" not in data:
            raise ValueError("Data does not contain X key")

        if "Y" not in data:
            raise ValueError("Data does not contain Y key")

        X = data["X"]
        Y = data["Y"]

        check = len(X.shape)
        if check != 2:
            raise ValueError(f"X must be 2-dim (not {check})")

        check = len(Y.shape)
        if len(Y.shape) != 3:
            raise ValueError(f"Y must be 3-dim (not {check})")


class DataGenFat(DataGenTorch[DataOut, DataLoader]):
    """
    Generates data for estimating fat shattering dimension.

    :param prior: Data generator for prior X.
    :type prior: DataGenTorch
    :param Sb: The number of binary samples to check shattering.
        Defaults to 10.
    :type Sb: int, optional
    :param Sr: The number of level offset samples to check shattering.
        Defaults to 10.
    :type Sr: int, optional
    :param gamma: The fat shattering parameter gamma.
        Defaults to 0.0 (pseudo-dimension).
    :type gamma: float, optional
    :param kwargs: Keyword arguments passed to the base class.
    """

    def __init__(
        self,
        prior: PriorTorch,
        Sb: int = 10,
        Sr: int = 10,
        gamma: float = 0.0,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.Sb = Sb
        self.Sr = Sr
        self.gamma = gamma
        self.prior = prior

    def gen_data(self, d: int) -> DataOut:
        """
        Generate the data.

        :param d: The number of inputs to shatter.
        :type d: int
        :return: A dictionary containing the generated data.
        :rtype: DataOut
        """

        X = self.prior.gen_data(d)
        b = generate_samples_b_fat(d=d, S=self.Sb, seed=self.seed)
        r = generate_samples_r_fat(d=d, S=self.Sr, seed=self.seed)
        Y = gen_synthetic_labels_fat(b, r, self.gamma, self.device)

        data = {"X": X, "Y": Y, "b": b, "r": r}

        return data

    def data_to_loader(self, data: DataOut, sr: int, sb: int) -> DataLoader:
        """
        Convert data to pytorch loader.

        :param data: Output of gen_data.
        :type data: DataOut
        :param sr: Current r sample.
        :type sr: int
        :param sb: Current b sample.
        :type sb: int
        :returns: Pytorch data loader.
        :rtype: DataLoader
        :raises ValueError: For invalid data or indeces sr or sb.
        """
        self._check_data(data)

        Nr = data["Y"].size(0)
        Nb = data["Y"].size(1)
        if sr < 0 or sr >= Nr:
            raise ValueError(f"Index sr ({sr}) must be between 0 and Ny ({Nr}).")
        if sb < 0 or sb >= Nb:
            raise ValueError(f"Index sb ({sb}) must be between 0 and Ny ({Nb}).")

        X = data["X"]
        Nx = X.size(0)
        Y = data["Y"][sr, sb]
        dataset = TensorDataset(X, Y)
        batch_size = Nx
        if self.batch_size is not None:
            batch_size = self.batch_size

        loader = DataLoader(dataset, batch_size=batch_size, shuffle=self.shuffle)

        return loader

    def _check_data(self, data: DataOut):
        if "X" not in data:
            raise ValueError("Data does not contain X key")

        if "Y" not in data:
            raise ValueError("Data does not contain Y key")

        X = data["X"]
        Y = data["Y"]

        check = len(X.shape)
        if check != 2:
            raise ValueError(f"X must be 2-dim (not {check})")

        check = len(Y.shape)
        if check != 4:
            raise ValueError(f"Y must be 4-dim (not {check})")


class DataGenRademacher(DataGenTorch[DataOut, DataLoader]):
    """
    Generates uniform data for estimating the empirical Rademacher complexity.

    :param prior: Prior for generating X samples.
    :type prior: PriorTorch
    :param num_sigma_samples: Number of samples for sigma.
            Defaults to 10.
    :type num_sigma_samples: int, optional
    :param num_data_samples: Number of samples for data sets.
            Defaults to 10.
    :type num_data_samples: int, optional
    :param gamma: The fat shattering parameter gamma
        Defaults to 0.0 (pseudo-dimension).
    :type gamma: float, optional
    :param scale: The re-scaling factor for uniform random numbers in [0, 1].
        Defaults to 2.0.
    :type scale: float, optional
    :param shift: The shift value for uniform random numbers in [0, 1].
        Defaults to -1.0.
    :type shift: float, optional
    :param kwargs: Keyword arguments passed to the base class.
    """

    def __init__(
        self,
        prior: PriorTorch,
        num_sigma_samples: int = 10,
        num_data_samples: int = 10,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.prior = prior
        self.num_sigma_samples = num_sigma_samples
        self.num_data_samples = num_data_samples

    def gen_data(self, m: int) -> DataOut:
        """Generate the data.

        :param m: Size of data set.
        :type m: int
        :param m: int:
        :return: A dictionary containing the generated data.
        :rtype: DataOut
        """

        X = self.prior.gen_data(m * self.num_data_samples)
        X = torch.reshape(X, (self.num_data_samples, m, self.prior.sizex))
        sigmas = gen_sigmas(m=m * self.num_sigma_samples, seed=self.seed, device=self.device)
        sigmas = torch.reshape(sigmas, (self.num_sigma_samples, m))

        data = {"X": X, "sigmas": sigmas}

        return data

    def data_to_loader(self, data: DataOut, s: int) -> DataLoader:
        """
        Convert data to pytorch loader.

        :param data: Output of gen_data.
        :type data: DataOut
        :param s: Current sample.
        :type s: int
        :return: Pytorch data loader.
        :rtype: DataLoader
        :raises ValueError: For invalid data or index s.
        """
        self._check_data(data)

        Ns = data["X"].size(0)
        if s < 0 or s >= Ns:
            raise ValueError(f"Index s ({s}) must be between 0 and Ns ({Ns}).")

        X = data["X"][s]
        Nx = X.size(0)
        Y = torch.zeros(Nx, device=X.device)
        dataset = TensorDataset(X, Y)
        batch_size = Nx
        if self.batch_size is not None:
            batch_size = self.batch_size

        loader = DataLoader(dataset, batch_size=batch_size, shuffle=self.shuffle)

        return loader

    def _check_data(self, data: DataOut):
        if "X" not in data:
            raise ValueError("Data does not contain X key")

        X = data["X"]
        check = len(X.shape)

        if check != 3:
            raise ValueError(f"X must be 3-dim (not {check})")


class UniformPrior(PriorTorch[Tensor]):
    """
    Generates uniform prior X.

    :param sizex: The size of the input data (dim of feature space).
    :type sizex: int
    :param scale: The re-scaling factor for uniform random numbers in [0, 1].
        Defaults to 2.0.
    :type scale: float, optional
    :param shift: The shift value for uniform random numbers in [0, 1].
        Defaults to -1.0.
    :type shift: float, optional
    :param kwargs: Keyword arguments passed to the base class.
    """

    def __init__(self, sizex: int, scale: float = 2.0, shift: float = -1.0, **kwargs) -> None:
        super().__init__(sizex, **kwargs)

        self.scale = scale
        self.shift = shift

    def gen_data(self, m: int) -> Tensor:
        """
        Generate the data.

        :param m: Size of data set.
        :type m: int
        :return: Prior X.
        :rtype: Tensor
        """

        X = gen_synthetic_features(
            d=m,
            sizex=self.sizex,
            seed=self.seed,
            scale=self.scale,
            shift=self.shift,
            device=self.device,
        )

        return X


class NormalPrior(PriorTorch[Tensor]):
    """
    Generates normal prior for X.

    :param sizex: The size of the input data (dim of feature space).
    :type sizex: int
    :param scale: The re-scaling factor for standard normal.
        Defaults to 1.0.
    :type scale: float, optional
    :param shift: The shift value for standard normal.
            Defaults to 0.0.
    :type shift: float, optional
    :param kwargs: Keyword arguments passed to the base class.
    """

    def __init__(self, sizex: int, scale: float = 1.0, shift: float = 0.0, **kwargs) -> None:
        super().__init__(sizex, **kwargs)

        self.scale = scale
        self.shift = shift

    def gen_data(self, m: int) -> Tensor:
        """
        Generate the data.

        :param m: Size of data set.
        :type m: int
        :return: Prior X.
        :rtype: Tensor
        """

        X = gen_synthetic_features_normal(
            d=m,
            sizex=self.sizex,
            seed=self.seed,
            scale=self.scale,
            shift=self.shift,
            device=self.device,
        )

        return X


def gen_dataset_capacity(
    N: int,
    sizex: int,
    num_samples: int = 10,
    seed: Optional[int] = None,
    scale: float = 2.0,
    shift: float = -1.0,
    device: Device = torch.device("cpu"),
) -> Tuple[Tensor, Tensor]:
    """
    Generates a dataset of input features x and output labels y.

    :param N: The number of inputs.
    :type N: int
    :param sizex: The dimension of each input.
    :type sizex: int
    :param num_samples: The number of output samples to generate. Defaults to 10.
    :type num_samples: int, optional
    :param seed: The random seed to use for generating the dataset.
        Defaults to None.
    :type seed: int, optional
    :param scale: The re-scaling factor for uniform random numbers in
        [0,1]. Defaults to 2.0.
    :type scale: float, optional
    :param shift: The shift value for uniform random numbers [0,1].
        Defaults to -1.0.
    :type shift: float, optional
    :param device: Torch device to run on. Defaults to CPU.
    :type device: Device
    :return: A tuple containing the input tensor x of shape (N, sizex)
        and the output (num_samples, N).
    :rtype: Tuple[Tensor, Tensor]
    """

    seed_ = seed
    if seed_ is None:
        seed_ = torch.seed()

    generator = torch.manual_seed(seed_)

    x = (
        scale
        * torch.rand(
            N,
            sizex,
            dtype=torch.float64,
            device=device,
            requires_grad=False,
            generator=generator,
        )
        + shift
    )
    y = (
        scale
        * torch.rand(
            num_samples,
            N,
            1,
            dtype=torch.float64,
            device=device,
            requires_grad=False,
            generator=generator,
        )
        + shift
    )

    return x, y


def generate_samples_b_fat(
    d: int,
    S: int,
    seed: Optional[int] = None,
) -> Array:
    """
    Generate S unique samples of b from {0, 1}^d.

    :param d: Number of input data samples for shattering.
    :type d: int
    :param S: Number of binary samples to check shattering (for
        scalability if d is large).
    :type S: int
    :param seed: The random seed. Defaults to None.
    :type seed: int, optional
    :return: An array of shape (S, d) containing unique samples of b.
    :rtype: Array
    """

    rng = np.random.default_rng(seed)
    max_values = 2**d
    if S >= max_values:
        possible_values = list(product([0, 1], repeat=d))
        return np.array(possible_values)
    else:
        samples: Set[Tuple[int, ...]] = set()
        while len(samples) < S:
            b_sample = tuple(rng.integers(0, 2, size=d))
            samples.add(b_sample)
        return np.array(list(samples))


def generate_samples_r_fat(
    d: int,
    S: int,
    seed: Optional[int] = None,
) -> Array:
    """
    Generate S samples of r from [0, 1]^d using Latin Hypercube Sampling.

    :param d: Dimension of the feature space.
    :type d: int
    :param S: Number of samples to generate.
    :type S: int
    :param seed: The random seed. Defaults to None.
    :type seed: int, optional
    :return: An array of shape (S, d) containing samples of r.
    :rtype: Array
    """

    sampler = qmc.LatinHypercube(d=d, seed=seed)
    r_samples = sampler.random(n=S)
    return r_samples


def gen_synthetic_features(
    d: int,
    sizex: int,
    seed: Optional[int] = None,
    scale: float = 2.0,
    shift: float = -1.0,
    device: Device = torch.device("cpu"),
) -> Tensor:
    """
    Generates d inputs x of dimension sizex sampled uniformly from scale*[0,1]+shift.

    :param d: The number of inputs x to generate.
    :type d: int
    :param sizex: The size of each input.
    :type sizex: int
    :param seed: The random seed to use for generating the features.
        Defaults to None.
    :type seed: int, optional
    :param scale: The re-scaling factor for uniform random numbers in
        [0,1]. Defaults to 2.0.
    :type scale: float, optional
    :param shift: The shift value for uniform random numbers [0,1].
        Defaults to -1.0.
    :type shift: float, optional
    :param device: Torch device to run on. Defaults to CPU.
    :type device: Device
    :return: Tensor X of shape (d, sizex).
    :rtype: Tensor
    """

    seed_ = seed
    if seed_ is None:
        seed_ = torch.seed()

    generator = torch.manual_seed(seed_)

    x = (
        scale
        * torch.rand(
            d,
            sizex,
            dtype=torch.float64,
            device=device,
            requires_grad=False,
            generator=generator,
        )
        + shift
    )

    return x


def gen_synthetic_features_normal(
    d: int,
    sizex: int,
    seed: Optional[int] = None,
    scale: float = 1.0,
    shift: float = 0.0,
    device: Device = torch.device("cpu"),
) -> Tensor:
    """
    Generates d inputs x of dimension sizex sampled from N(shift, scale^2).

    :param d: The number of inputs x to generate.
    :type d: int
    :param sizex: The size of each input.
    :type sizex: int
    :param seed: The random seed to use for generating the features.
        Defaults to None.
    :type seed: int, optional
    :param scale: The re-scaling factor for uniform random numbers in
        [0,1]. Defaults to 1.0.
    :type scale: float, optional
    :param shift: The shift value for uniform random numbers [0,1].
        Defaults to 0.0.
    :type shift: float, optional
    :param device: Torch device to run on. Defaults to CPU.
    :type device: Device, optional
    :return: Tensor X of shape (d, sizex).
    :rtype: Tensor
    """

    seed_ = seed
    if seed_ is None:
        seed_ = torch.seed()

    generator = torch.manual_seed(seed_)

    x = (
        scale
        * torch.randn(
            d,
            sizex,
            dtype=torch.float64,
            device=device,
            requires_grad=False,
            generator=generator,
        )
        + shift
    )

    return x


def gen_synthetic_labels_fat(
    b: Array,
    r: Array,
    gamma: float = 0.0,
    device: Device = torch.device("cpu"),
) -> Tensor:
    """
    Generate constant label values equal to r_i + gamma
    when b_i = 1 and r_i - gamma when b_i = 0.

    :param b: An array of shape (Sb,d)
            containing Sb samples of d-dim binary values.
    :type b: Array
    :param r: An array of shape (Sr,d)
            containing Sr samples of d-dim real values in [0, 1].
    :type r: Array
    :param gamma: The fat-shattering margin value.
            Defaults to 0.0.
    :type gamma: float, optional
    :param device: Torch device to run on. Defaults to CPU.
    :type device: Device, optional
    :return: Y of shape (Sr, Sb, d, 1).
    :rtype: Tensor
    :raises ValueError: If the length of b[0] is not the same as the length of r[0].
    """

    Sb = len(b)
    Sr = len(r)
    d = d1 = len(b[0])
    d2 = len(r[0])

    if d1 != d2:
        raise ValueError(
            f"The length of b[0] and r[0] are {d1} and {d2}. " f"Should be constant and the same."
        )

    labels = np.zeros((Sr, Sb, d, 1))
    for sr in range(Sr):
        for sb in range(Sb):
            for i in range(d):
                if b[sb][i] == 1:
                    labels[sr, sb, i, 0] = r[sr, i] + gamma
                else:
                    labels[sr, sb, i, 0] = r[sr, i] - gamma

    y = torch.tensor(labels, dtype=torch.float64, device=device, requires_grad=False)

    return y


def gen_sigmas(m: int, seed: Optional[int] = None, device: Device = torch.device("cpu")) -> Tensor:
    """
    Random vector of +-1.

    :param m: Number of sigmas.
    :type m: int
    :param seed: The random seed to use for generating the features.
        Defaults to None.
    :type seed: int, optional
    :param device: Torch device to run on. Defaults to CPU.
    :type device: Device, optional
    :return: Tensor of sigmas.
    :rtype: Tensor
    """

    seed_ = seed
    if seed_ is None:
        seed_ = torch.seed()

    generator = torch.manual_seed(seed_)

    sigmas = torch.randint(2, (m,), device=device, requires_grad=False, generator=generator) * 2 - 1

    return sigmas


def generate_lhs_samples(
    n_samples: int,
    n_dims: int,
    lower_bound: float,
    upper_bound: float,
    seed: Optional[int] = None,
) -> Array:
    """
    Generate Latin Hypercube Samples (LHS) within the specified range.

    :param n_samples: The number of samples to generate.
    :type n_samples: int
    :param n_dims: The number of dimensions for each sample.
    :type n_dims: int
    :param lower_bound: The lower bound for the samples.
    :type lower_bound: float
    :param upper_bound: The upper bound for the samples.
    :type upper_bound: float
    :param seed: The random seed. Defaults to None.
    :type seed: int, optional
    :return: A numpy array containing the generated samples.
    :rtype: Array
    """

    sampler = qmc.LatinHypercube(d=n_dims, seed=seed)
    sample = sampler.random(n=n_samples)
    return lower_bound + sample * (upper_bound - lower_bound)


def generate_model_lhs_samples(
    model: Model,
    n_samples: int,
    lower_bound: float,
    upper_bound: float,
    device: Optional[CDevice] = None,
    dtype: Optional[DType] = None,
    seed: Optional[int] = None,
) -> ParameterList:
    """
    Generate Latin Hypercube Samples (LHS) for each parameter of a given model.

    :param model: The model whose parameters to sample.
    :type model: Model
    :param n_samples: The number of samples to generate.
    :type n_samples: int
    :param lower_bound: The lower bound for the samples.
    :type lower_bound: float
    :param upper_bound: The upper bound for the samples.
    :type upper_bound: float
    :param cdevice: Classical device to store the parameters, defaults to None.
    :type cdevice: CDevice, optional
    :param dtype: Data type of the parameters, defaults to None.
    :type dtype: DType, optional
    :param seed: The random seed. Defaults to None.
    :type seed: int, optional
    :return: A list of lists containing samples for each parameter.
    :rtype: ParameterList
    """

    model_parameters = list(filter(lambda p: p.requires_grad, model.parameters()))
    samples = []
    for p in model_parameters:
        n_dims = p.numel()
        sample_parameter = generate_lhs_samples(
            n_samples=n_samples,
            n_dims=n_dims,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
            seed=seed,
        )
        samples.append(sample_parameter.reshape((n_samples,) + p.shape))
    parameter_list = [
        [torch.tensor(samples[j][i], device=device, dtype=dtype) for j in range(len(samples))]
        for i in range(n_samples)
    ]

    return parameter_list



import logging

import torch

from .datagen import DataGenFat as Datagen
from .trainer import SupervisedTrainer as Trainer
from .types import QModel, Tensor


def fat_shattering_dim(
    model: QModel,
    datagen: Datagen,
    trainer: Trainer,
    dmin: int,
    dmax: int,
    gamma: float = 0.0,
    dstep: int = 1,
) -> int:
    """
    Estimate the fat-shattering dimension for a model with a given architecture.

    :param model: The model.
    :type model: QModel
    :param datagen: The (synthetic) data generator.
    :type datagen: Datagen
    :param trainer: The trainer.
    :type trainer: Trainer
    :param dmin: Iteration start for dimension check.
    :type dmin: int
    :param dmax: Iteration stop for dimension check (including).
    :type dmax: int
    :param gamma: The margin value. Defaults to 0.0 (pseudo-dim).
    :type gamma: float, optional
    :param dstep: Dimension iteration step size. Defaults to 1.
    :type dstep: int
    :return: The estimated fat-shattering dimension.
    :rtype: int
    """

    for d in range(dmin, dmax + 1, dstep):
        shattered = check_shattering(model, datagen, trainer, d, gamma)

        if not shattered:
            if d == dmin:
                logging.basicConfig(level=logging.WARNING)
                logging.warning(f"Stopped at dmin = {dmin}.")
                return dmin

            return d - dstep

    logging.basicConfig(level=logging.WARNING)
    logging.warning(f"Reached dmax = {dmax}.")
    return dmax


def check_shattering(
    model: QModel, datagen: Datagen, trainer: Trainer, d: int, gamma: float
) -> bool:
    """
    Check if the model shatters a given dimension d with margin value gamma.

    :param model: The model.
    :type model: QModel
    :param datagen: The (synthetic) data generator.
    :type datagen: Datagen
    :param trainer: The trainer.
    :type trainer: Trainer
    :param d: Size of data set to shatter.
    :type d: int
    :param gamma: The margin value.
    :type gamma: float
    :return: True if the model shatters a random data set of size d, False otherwise.
    :rtype: bool
    """

    data = datagen.gen_data(d)
    X = data["X"]
    b = data["b"]
    r = data["r"]

    for sr in range(len(r)):
        shattered = True
        for sb in range(len(b)):
            loader = datagen.data_to_loader(data, sr, sb)
            trainer.train(model, loader, loader)
            predictions = model(X)

            for i, pred in enumerate(predictions):
                if b[sb, i] == 1 and not (pred >= r[sr, i] + gamma):
                    shattered = False
                    break
                if b[sb, i] == 0 and not (pred <= r[sr, i] - gamma):
                    shattered = False
                    break

            if not shattered:
                break

        if shattered:
            return True

    return False


def normalize_const(weights: Tensor, gamma: float, Rx: float) -> float:
    """
    Compute a normalization constant given a tensor of weights and
    the margin parameter gamma.

    Rationale: the fat-shattering dimension of a linear classifier,
    with weights bounded by Rw and data bounded by Rx, is bounded by
    <= Rw^2*Rx^2/gamma^2. Hence, normalizing the fat-shattering dimension
    of a model with unbounded weights compares it to the best linear classifier
    with the same weight norm.

    :param weights: Tensor of weights
    :type weights: Tensor
    :param gamma: Margin parameter.
    :type gamma: float
    :param Rx: Estimated 2-radius of input data.
    :type Rx: float
    :return: A positive real-valued normalization constant.
    :rtype: float
    """

    V = torch.norm(weights, p=2)
    V = V.item()
    C = V**2 * Rx**2 / gamma**2

    return C



import math
from typing import List

import tntorch
import torch

from qulearn.hat_basis import HatBasis

from .types import MPS, Tensor


class MPSQGates:
    """
    Converts Matrix Product States (MPS) to quantum gates.

    :param mps: The MPS from which quantum gates will be extracted.
    :type mps: MPS
    """

    def __init__(self, mps: MPS) -> None:
        self.mps = mps
        self.max_rank_power = compute_max_rank_power(mps)

    def qgates(self) -> List[Tensor]:
        """
        Extracts quantum gates from the MPS.

        :returns: A list of Tensors, each representing a unitary matrix for a quantum gate.
        :rtype: List[Tensor]
        """

        N = self.mps.dim()
        mps = self.pad_cores()
        mps.orthogonalize(N - 1)

        Us = []
        core = self.contract(mps=mps, L=self.max_rank_power + 1)
        Q0 = self.left_core_reshape(core)
        Us.append(embed2unitary(Q0))

        cores = mps.cores.copy()
        for j in range(self.max_rank_power + 1, N):
            core = cores[j]
            Q = self.reg_core_reshape(core)
            Us.append(embed2unitary(Q))

        return Us

    def pad_cores(self) -> MPS:
        """
        Pads the cores of the MPS to match the maximum rank, ensuring uniform core dimensions.

        :returns: A new MPS with padded cores.
        :rtype: MPS
        """

        chi = 2**self.max_rank_power
        cores = self.mps.cores.copy()

        core = cores[0]
        rankR = chi - core.shape[-1]
        pad_first = (0, rankR)
        cores[0] = torch.nn.functional.pad(core, pad_first)

        core = cores[-1]
        rankL = chi - core.shape[0]
        pad_last = (0, 0, 0, 0, 0, rankL)
        cores[-1] = torch.nn.functional.pad(core, pad_last)

        for j in range(1, len(cores) - 1):
            core = cores[j]
            rankL = chi - core.shape[0]
            rankR = chi - core.shape[-1]
            pad = (0, rankR, 0, 0, 0, rankL)
            cores[j] = torch.nn.functional.pad(core, pad)

        mps = tntorch.Tensor(cores)

        return mps

    def contract(self, mps, L: int) -> Tensor:
        """
        Contracts the first L cores from the left of the MPS to form a single tensor.

        :param mps: The MPS from which quantum gates will be extracted.
        :type mps: MPS
        :param L: The number of cores to contract.
        :type L: int
        :returns: A Tensor resulting from the contraction of the first L cores.
        :rtype: Tensor
        """

        cores = mps.cores
        result = cores[0].clone().detach()
        for i in range(1, L):
            result = torch.tensordot(result, cores[i], dims=([result.dim() - 1], [0]))

        return result

    def left_core_reshape(self, core: Tensor) -> Tensor:
        """
        Reshapes a core tensor for the left-most core,
        preparing it for SVD and embedding into a unitary matrix.

        :param core: The core tensor to reshape.
        :type core: Tensor
        :returns: The reshaped core tensor.
        :rtype: Tensor
        """

        rows = list(range(0, self.max_rank_power + 2))
        row_size = int(torch.prod(torch.tensor(core.shape)[rows]).item())
        Q = core.reshape(row_size, -1)
        return Q

    def reg_core_reshape(self, core: Tensor) -> Tensor:
        """
        Reshapes a regular core tensor, preparing it for SVD and embedding into a unitary matrix.

        :param core: The core tensor to reshape.
        :type core: Tensor
        :returns: The reshaped core tensor.
        :rtype: Tensor
        """

        rows = [0, 1]
        row_size = int(torch.prod(torch.tensor(core.shape)[rows]).item())
        Q = core.reshape(row_size, -1)
        return Q


class HatBasisMPS:
    """
    Generates Matrix Product States (MPS) corresponding to evaluations
    of linear hat basis functions.

    :param basis: The hat basis to use for generating the MPS.
    :type basis: HatBasis

    .. note::
       The number of nodes in the hat basis must be a power of 2,
       corresponding to the number of qubits used.
       Currently, works only for scalar inputs x.
    """

    def __init__(self, basis: HatBasis) -> None:
        self.basis = basis

        num_qubits = math.log2(basis.num_nodes)
        if not num_qubits.is_integer():
            raise ValueError(f"Number of nodes ({basis.num_nodes}) " "must be a power of 2.")

        self.num_sites = int(num_qubits)

    def __call__(self, x: Tensor) -> MPS:
        """
        Constructs the MPS of the hat basis evaluated at a given point x.

        :param x: The input at which to evaluate the hat basis.
        :type x: Tensor
        :returns: The MPS at point x.
        :rtype: MPS
        """

        return self.eval(x)

    def eval(self, x: Tensor) -> MPS:
        """
        Constructs the MPS of the hat basis evaluated at a given point x.

        :param x: The input at which to evaluate the hat basis.
        :type x: Tensor
        :returns: The MPS at point x.
        :rtype: MPS
        """

        position = self.basis.position(x)
        first, second = self.basis.nonz_vals(x)

        # works only for scalars
        idx = int(position.item())
        a = first.item()
        b = second.item()

        mps = self.mps_hatbasis(a, b, idx)

        return mps

    def mps_hatbasis(self, first: float, second: float, idx: int) -> MPS:
        """
        Generates an MPS for the hat basis vector.

        :param first: The first non-zero value in the hat basis function.
        :type first: float
        :param second: The second non-zero value in the hat basis function.
        :type second: float
        :param idx: The index of the hat basis function.
        :type idx: int
        :returns: An MPS representing the hat basis.
        :rtype: MPS
        """

        even = idx % 2 == 0
        if even:
            mps = self.mps_hatbasis_evenidx(first, second, idx)
        else:
            mps = self.mps_hatbasis_oddidx(first, second, idx)

        return mps

    def mps_hatbasis_evenidx(self, first: float, second: float, idx: int) -> MPS:
        """
        Generates an MPS for an even index in the hat basis.

        :param first: The first non-zero value in the hat basis function.
        :type first: float
        :param second: The second non-zero value in the hat basis function.
        :type second: float
        :param idx: The index of the hat basis function.
        :type idx: int
        :returns: An MPS representing the hat basis.
        :rtype: MPS
        """
        binidx = format(idx, "0{}b".format(self.num_sites))

        cores = []
        for j in range(self.num_sites - 1):
            k = int(binidx[j])
            core = torch.zeros((1, 2, 1))
            core[0, k, 0] = 1.0
            cores.append(core)

        core = torch.zeros(1, 2, 1)
        core[0, 0, 0] = first
        core[0, 1, 0] = second
        cores.append(core)
        mps = tntorch.Tensor(cores)

        return mps

    def mps_hatbasis_oddidx(self, first: float, second: float, idx: int) -> MPS:
        """
        Generates an MPS for an odd index in the hat basis.

        :param first: The first non-zero value in the hat basis function.
        :type first: float
        :param second: The second non-zero value in the hat basis function.
        :type second: float
        :param idx: The index of the hat basis function.
        :type idx: int
        :returns: An MPS representing the hat basis.
        :rtype: MPS
        """

        binidx = format(idx, "0{}b".format(self.num_sites))
        zerobit = zerobit_position_odd(idx, self.num_sites)

        cores = []
        for j in range(zerobit):
            k = int(binidx[j])
            core = torch.zeros((1, 2, 1))
            core[0, k, 0] = 1.0
            cores.append(core)

        if zerobit != self.num_sites - 1:
            core = torch.zeros(1, 2, 2)
            core[0, 0, 1] = 1.0
            core[0, 1, 0] = 1.0
            cores.append(core)

        for j in range(zerobit + 1, self.num_sites - 1):
            core = torch.zeros((2, 2, 2))
            core[0, 0, 0] = 1.0
            core[1, 1, 1] = 1.0
            cores.append(core)

        if zerobit != self.num_sites - 1:
            core = torch.zeros(2, 2, 1)
            core[0, 0, 0] = second
            core[1, 1, 0] = first
            cores.append(core)
        else:
            core = torch.zeros(1, 2, 1)
            core[0, 0, 0] = second
            core[0, 1, 0] = first
            cores.append(core)

        mps = tntorch.Tensor(cores)

        return mps


def compute_max_rank_power(mps: MPS) -> int:
    """
    Computes the power in base 2 of the maximum rank of an MPS tensor.

    :param mps: An MPS tensor whose maximum rank is to be computed.
    :type mps: MPS
    :return: The maximum rank power as an integer.
    :rtype: int
    """

    s = int(torch.ceil(torch.log2(max(mps.ranks_tt))))
    return s


def embed2unitary(Q: Tensor) -> Tensor:
    """
    Embeds a matrix Q into a unitary matrix by appending orthonormal columns.

    :param Q: The input matrix to be embedded into a unitary matrix.
              Its shape is assumed to be (m, k), where m >= k.
    :type Q: Tensor
    :return: The unitary matrix formed by appending orthonormal columns to Q.
    :rtype: Tensor
    """

    k = Q.shape[-1]
    U, _, _ = torch.linalg.svd(Q, full_matrices=True)
    Q_ = torch.cat((Q, U[:, k:]), dim=1)
    return Q_


def zerobit_position_odd(k, n):
    """
    Finds the position of the most least zero bit in the binary representation of an odd integer.

    :param k: An odd integer.
    :type k: int
    :param n: The total length for the binary representation.
    :type n: int
    :return: The position of the least significant zero bit.
    :rtype: int
    """

    if k % 2 == 0:
        m = 0
    else:
        m = int(math.log2((-k) & (k + 1)))
    pos = n - m - 1
    return int(pos)



from typing import Optional

import torch

from .types import Loss, Tensor


class RademacherLoss(Loss):
    """
    Computes the loss function required to estimate emprical Rademacher complexity.

    :param sigmas: 1D Tensor of sigmas.
    :type sigmas: Tensor
    """

    def __init__(self, sigmas: Tensor) -> None:
        super().__init__()

        self._check_sigmas(sigmas)
        self.sigmas = sigmas

    def forward(self, output: Tensor, _: Optional[Tensor] = None) -> Tensor:
        """
        Compute loss = -1/m * sum_k sigma_k*f_k.

        :param output: Predicted value tensor.
        :type output: Tensor
        :param _: Ignored. Defaults to None.
        :type _: Tensor
        :return: Loss value.
        :rtype: Tensor
        :raises ValueError: If output not a 1D Tensor, or not the same length as sigmas.
        """

        len_shape = len(output.shape)
        if len_shape != 2:
            raise ValueError(f"output (dim = {len_shape}) should be a 2D tensor")
        second_dim = output.shape[1]
        if second_dim != 1:
            raise ValueError(f"output second dim should be 1 not {second_dim}")

        len_out = output.shape[0]
        len_sigmas = self.sigmas.shape[0]
        if len_out != len_sigmas:
            raise ValueError(
                f"output length ({len_out}) does not match sigmas length ({len_sigmas})"
            )

        loss = -(self.sigmas * output[:, 0]).mean()

        return loss

    def _check_sigmas(self, sigmas: Tensor) -> None:
        """
        Ensures sigmas has the expected format.

        :param sigmas: sigmas Tensor.
        :type sigmas: Tensor
        :return: None
        :raises ValueError: If sigmas has invalid shape or values.
        """

        len_shape = len(sigmas.shape)

        if len_shape != 1:
            raise ValueError(f"sigmas (dim={len_shape}) should be a 1D tensor")

        if not torch.all((sigmas == 1) | (sigmas == -1)):
            raise ValueError("All sigmas must be 1 or -1")



import tntorch
import torch

from .types import MPS


def kron(tleft: MPS, tright: MPS) -> MPS:
    """
    Performs the Kronecker product of two MPS tensors.

    :param tleft: The first MPS tensor.
    :type tleft: MPS
    :param tright: The second MPS tensor.
    :type tright: MPS
    :return: The MPS tensor resulting from the Kronecker product of `tleft` and `tright`.
    :rtype: MPS
    """
    c1 = tleft.cores
    c2 = tright.cores
    c3 = c1 + c2
    t3 = tntorch.Tensor(c3)

    return t3


def zkron(tleft: MPS, tright: MPS) -> MPS:
    """
    Performs the z-ordered Kronecker product of two MPS tensors.
    See https://arxiv.org/abs/1802.02839.

    :param tleft: The first MPS tensor.
    :type tleft: MPS
    :param tright: The second MPS tensor.
    :type tright: MPS
    :return: The MPS tensor resulting from the Kronecker product of `tleft` and `tright`.
    :rtype: MPS
    """
    _core_length_check(tleft, tright)

    coresleft = tleft.cores
    coresright = tright.cores

    if len(coresleft) != len(coresright):
        raise ValueError("The number of cores in the left and right MPS must be the same.")

    coresout = []

    for i in range(len(coresleft)):
        coreleft = coresleft[i]
        coreright = coresright[i]
        rankleft1 = coreleft.shape[0]
        rankleft2 = coreleft.shape[-1]
        rankright1 = coreright.shape[0]
        rankright2 = coreright.shape[-1]

        site_dim = coreleft.shape[1]
        core = torch.empty((rankleft1 * rankright1, site_dim, rankleft2 * rankright1))
        for k in range(site_dim):
            core[:, k, :] = torch.kron(coreleft[:, k, :], torch.eye(rankright1))
        coresout.append(core)

        site_dim = coreright.shape[1]
        core = torch.empty((rankleft2 * rankright1, site_dim, rankleft2 * rankright2))
        for k in range(site_dim):
            core[:, k, :] = torch.kron(torch.eye(rankleft2), coreright[:, k, :])
        coresout.append(core)

    tout = tntorch.Tensor(coresout)
    return tout


def zkron_joined(tleft, tright):
    """
    Performs the z-ordered Kronecker product of two MPS tensors,
    and joins the physical indices of tleft and tright into one.
    See https://arxiv.org/abs/1802.02839.

    :param tleft: The first MPS tensor.
    :type tleft: MPS
    :param tright: The second MPS tensor.
    :type tright: MPS
    :return: The MPS tensor resulting from the Kronecker product of `tleft` and `tright`.
    :rtype: MPS
    """
    _core_length_check(tleft, tright)

    c1 = tleft.cores
    c2 = tright.cores
    c3 = [torch.kron(a, b) for a, b in zip(c1, c2)]

    t3 = tntorch.Tensor(c3)
    return t3


def _core_length_check(tleft, tright):
    coresleft = tleft.cores
    coresright = tright.cores
    if len(coresleft) != len(coresright):
        raise ValueError("The number of cores in the left and right MPS must be the same.")



import math
from enum import Enum
from typing import Dict, Iterable, Optional

import pennylane as qml
import torch
from torch import nn

from .hat_basis import HatBasis
from .mps import HatBasisMPS, MPSQGates
from .mps_kronprod import kron, zkron
from .types import (
    CDevice,
    DType,
    Entropy,
    Expectation,
    Observable,
    Observables,
    Probability,
    QDevice,
    QNode,
    Sample,
    Tensor,
    Wires,
)

DEFAULT_QDEV_CFG = {"name": "default.qubit", "shots": None}


class MeasurementType(Enum):
    """Measurement type for a measurement layer."""

    """Expectation: return expected value of observable."""
    Expectation = "expectation"
    """Probabilities: return vector of probabilities."""
    Probabilities = "probabilities"
    """Samples: return measurement samples."""
    Samples = "samples"
    """Entropy: calculate the von Neumann entropy of a subsystem."""
    Entropy = "entropy"


class CircuitLayer(nn.Module):
    """
    Base class for a quantum circuit layer.

    A circuit layer transforms a quantum state but does not perform any measurements.
    Thus, a circuit layer produces no classical output.
    It can be combined with a measurement layer that returns classical output.
    By default, the base class does nothing (zero state).
    Derived classes need to override :meth:`circuit`.

    :param wires: Number or list of circuits.
    :type wires: Wires
    """

    def __init__(self, wires: Wires) -> None:
        super().__init__()
        self.set_wires(wires)

    def forward(self, x: Tensor) -> None:
        """Forward pass. See :meth:`circuit`"""
        self.circuit(x)

    def circuit(self, _: Tensor) -> None:
        """Applies input-dependent unitary transformations to a circuit.

        :param x: Input data samples.
        :type x: Tensor.
        :return: None for the base class.
        :rtype: None.
        """
        return None

    def set_wires(self, wires: Wires) -> None:
        """Set circuit (qubit) wires."""
        if isinstance(wires, int):
            self.num_wires = wires
            self.wires = list(range(wires))
        else:
            self.num_wires = len(list(wires))
            self.wires = list(wires)


class IQPEmbeddingLayer(CircuitLayer):
    """
    Layer for IQP (Instantaneous Quantum Polynomial) embedding.

    :param wires: The wires to be used by the layer
    :type wires: Wires
    :param n_repeat: The number of times to repeat the IQP embedding, defaults to 1
    :type n_repeat: int, optional
    :param kwargs: Extra arguments passed to the IQP embedding
    """

    def __init__(self, wires: Wires, n_repeat: int = 1, **kwargs) -> None:
        super().__init__(wires)
        self.n_repeat = n_repeat
        self.kwargs = kwargs
        self.qfunc = qml.IQPEmbedding

    def circuit(self, x: Tensor) -> None:
        """
        Define the quantum circuit for this layer.

        :param x: Input tensor that is passed to the quantum circuit.
        :type x: Tensor
        """
        for _ in range(self.n_repeat):
            self.qfunc(x, wires=self.wires, **self.kwargs)


class HatBasisQFE(CircuitLayer):
    """
    Layer for the 1D hat basis quantum feature embedding.

    :param basis: The hat basis class.
    :type basis: HatBasis
    :param wires: The wires to be used by the layer
    :type wires: Wires
    :param sqrt: Set flag to take square roots before applying hat basis.
    :type sqrt: bool
    :param normalize: Set flag to normalize basis vector before embedding.
    :type normalize: bool
    """

    def __init__(
        self,
        wires: Wires,
        basis: HatBasis,
        sqrt: bool = False,
        normalize: bool = False,
    ) -> None:
        super().__init__(wires)
        self.basis = basis
        self.sqrt = sqrt
        self.normalize = normalize
        self.norm = 1.0
        self.hbmps = HatBasisMPS(basis)

    def circuit(self, x: Tensor) -> None:
        """
        Define the quantum circuit for this layer.

        :param x: Input tensor that is passed to the quantum circuit.
        :type x: Tensor
        """

        position = int(self.basis.position(x))
        a, b = self.basis.nonz_vals(x)

        if self.sqrt:
            # sometimes the values are close to 0 and negative
            a = torch.sqrt(torch.abs(a))
            b = torch.sqrt(torch.abs(b))

        if position == -1:
            self.norm = b.item()
            for q in range(self.num_wires):
                qml.Identity(wires=self.wires[q])
            return None

        elif position == -2:
            self.norm = a.item()
            for q in range(self.num_wires):
                qml.PauliX(wires=self.wires[q])
            return None

        self.norm = torch.sqrt(a**2 + b**2).item()
        if self.normalize:
            a /= self.norm
            b /= self.norm

        # for compatibility (TODO: remove)
        first = a.item()
        second = b.item()

        mps = self.hbmps.mps_hatbasis(first, second, position)
        mpsgates = MPSQGates(mps)

        s = mpsgates.max_rank_power
        Us = mpsgates.qgates()
        N = len(Us)
        count = 0
        for k in range(N - 1, -1, -1):
            wires_idx = list(range(self.num_wires - count - s - 1, self.num_wires - count))
            subwires = [self.wires[idx] for idx in wires_idx]
            qml.QubitUnitary(Us[k], wires=subwires, unitary_check=False)

            count += 1

    def compute_norm(self, x: Tensor) -> float:
        """
        Compute the norm of the basis vector for the given input x.

        :param x: Input tensor that is passed to basis vector.
        :type x: Tensor
        :returns: The norm.
        :rtype: float
        """
        position = int(self.basis.position(x))
        a, b = self.basis.nonz_vals(x)

        if self.sqrt:
            a = torch.sqrt(a)
            b = torch.sqrt(b)

        if position == -1:
            self.norm = b.item()
            return self.norm

        elif position == -2:
            self.norm = a.item()
            return self.norm

        self.norm = torch.sqrt(a**2 + b**2).item()
        return self.norm


class Linear2DBasisQFE(CircuitLayer):
    """
    Layer for the 2D hat basis quantum feature embedding.

    :param basis: The 1D hat basis class.
    :type basis: HatBasis
    :param wires: The wires to be used by the layer
    :type wires: Wires
    :param sqrt: Set flag to take square roots before applying hat basis.
    :type sqrt: bool
    :param normalize: Set flag to normalize basis vector before embedding.
    :type normalize: bool
    """

    def __init__(
        self,
        wires: Wires,
        basis: HatBasis,
        sqrt: bool = False,
        normalize: bool = False,
        zorder: bool = False,
    ) -> None:
        super().__init__(wires)
        self.basis = basis
        self.sqrt = sqrt
        self.normalize = normalize
        self.norm = 1.0
        self.hbmps = HatBasisMPS(basis)
        self.zorder = zorder
        self.mps = None
        self.mps1 = None
        self.mps2 = None

    def circuit(self, x: Tensor) -> None:
        """
        Define the quantum circuit for this layer.

        :param x: Input tensor that is passed to the quantum circuit.
        :type x: Tensor
        """
        self._check_input(x)

        x1 = x[0]
        x2 = x[1]
        position1 = int(self.basis.position(x1))
        position2 = int(self.basis.position(x2))
        a1, b1 = self.basis.nonz_vals(x1)
        a2, b2 = self.basis.nonz_vals(x2)

        if self.sqrt:
            # sometimes the values are close to 0 and negative
            a1 = torch.sqrt(torch.abs(a1))
            b1 = torch.sqrt(torch.abs(b1))
            a2 = torch.sqrt(torch.abs(a2))
            b2 = torch.sqrt(torch.abs(b2))

        # TODO: cover the case where x or y are outside of bounds

        val1 = a1 * a2
        val2 = a1 * b2
        val3 = a2 * b1
        val4 = a2 * b2
        norm = torch.sqrt(val1**2 + val2**2 + val3**2 + val4**2)

        if self.normalize:
            a1 /= torch.sqrt(norm)
            b1 /= torch.sqrt(norm)
            a2 /= torch.sqrt(norm)
            b2 /= torch.sqrt(norm)

        self.norm = norm.item()

        # for compatibility (TODO: remove)
        first1 = a1.item()
        second1 = b1.item()
        first2 = a2.item()
        second2 = b2.item()

        mps1 = self.hbmps.mps_hatbasis(first1, second1, position1)
        mps2 = self.hbmps.mps_hatbasis(first2, second2, position2)

        if self.zorder:
            mps = zkron(mps2, mps1)
        else:
            mps = kron(mps2, mps1)

        self.mps1 = mps1
        self.mps2 = mps2
        self.mps = mps
        mpsgates = MPSQGates(mps)

        s = mpsgates.max_rank_power
        Us = mpsgates.qgates()
        N = len(Us)
        count = 0
        for k in range(N - 1, -1, -1):
            wires_idx = list(range(self.num_wires - count - s - 1, self.num_wires - count))
            subwires = [self.wires[idx] for idx in wires_idx]
            qml.QubitUnitary(Us[k], wires=subwires, unitary_check=False)

            count += 1

    def compute_norm(self, x: Tensor) -> float:
        """
        Compute the norm of the basis vector for the given input x.

        :param x: Input tensor that is passed to basis vector.
        :type x: Tensor
        :returns: The norm.
        :rtype: float
        """
        self._check_input(x)

        x1 = x[0]
        x2 = x[1]
        a1, b1 = self.basis.nonz_vals(x1)
        a2, b2 = self.basis.nonz_vals(x2)

        if self.sqrt:
            # sometimes the values are close to 0 and negative
            a1 = torch.sqrt(torch.abs(a1))
            b1 = torch.sqrt(torch.abs(b1))
            a2 = torch.sqrt(torch.abs(a2))
            b2 = torch.sqrt(torch.abs(b2))

        # TODO: cover the case where x or y are outside of bounds

        val1 = a1 * a2
        val2 = a1 * b2
        val3 = a2 * b1
        val4 = a2 * b2
        self.norm = torch.sqrt(val1**2 + val2**2 + val3**2 + val4**2).item()

        return self.norm

    def _check_input(self, x: Tensor):
        if x.dim() > 2:
            raise ValueError("Input tensor must have 2 dimensions")

        if torch.any(torch.abs(x) >= 1):
            raise ValueError("Out of bounds case is not implemented")


class RYCZLayer(CircuitLayer):
    """
    Layer for the RYCZ (Rotation around Y and Controlled-Z) gates.

    :param wires: The wires to be used by the layer.
    :type wires: Wires
    :param n_layers: The number of layers for the simplified two-design architecture,
        defaults to 1.
    :type n_layers: int, optional
    :param cdevice: Classical device to store the initial layer weights
        and internal layer weights, defaults to None.
    :type cdevice: CDevice, optional
    :param dtype: Data type of the weights, defaults to None.
    :type dtype: DType, optional
    :param kwargs: Extra arguments passed to the SimplifiedTwoDesign.
    """

    def __init__(
        self,
        wires: Wires,
        n_layers: int = 1,
        cdevice: Optional[CDevice] = None,
        dtype: Optional[DType] = None,
        **kwargs,
    ) -> None:
        super().__init__(wires)

        self.n_layers = n_layers
        self.qfunc = qml.SimplifiedTwoDesign
        self.cdevice = cdevice
        self.dtype = dtype
        self.kwargs = kwargs

        self.initial_layer_weights = torch.nn.Parameter(
            torch.empty(self.num_wires, device=self.cdevice, dtype=self.dtype)
        )
        self.weights = torch.nn.Parameter(
            torch.empty(
                (self.n_layers, self.num_wires - 1, 2),
                device=self.cdevice,
                dtype=self.dtype,
            )
        )
        nn.init.uniform_(self.initial_layer_weights, a=0.0, b=2 * math.pi)
        nn.init.uniform_(self.weights, a=0.0, b=2 * math.pi)

    def circuit(self, _: Optional[Tensor] = None) -> None:
        """
        Define the quantum circuit for this layer.

        :param _: Input tensor that is passed to the quantum circuit (ignored).
        :type x: Optional[Tensor]
        """
        self.qfunc(self.initial_layer_weights, self.weights, self.wires, **self.kwargs)


class AltRotCXLayer(CircuitLayer):
    """
    Layer for alternating CNOT gates with universal 1-qubit rotation.

    :param wires: The wires to be used by the layer.
    :type wires: Wires
    :param n_layers: The number of layers for the simplified two-design architecture,
        defaults to 1.
    :type n_layers: int, optional
    :param cdevice: Classical device to store the initial layer weights
        and internal layer weights, defaults to None.
    :type cdevice: CDevice, optional
    :param dtype: Data type of the weights, defaults to None.
    :type dtype: DType, optional
    :param kwargs: Extra arguments passed to the SimplifiedTwoDesign.
    """

    def __init__(
        self,
        wires: Wires,
        n_layers: int = 1,
        cdevice: Optional[CDevice] = None,
        dtype: Optional[DType] = None,
        **kwargs,
    ) -> None:
        super().__init__(wires)

        self.n_layers = n_layers
        self.cdevice = cdevice
        self.dtype = dtype
        self.kwargs = kwargs

        # weight parameters
        self.initial_layer_weights = torch.nn.Parameter(
            torch.empty((self.num_wires, 3), device=self.cdevice, dtype=self.dtype)
        )
        self.weights = torch.nn.Parameter(
            torch.empty(
                (self.n_layers, 2 * (self.num_wires - 1), 3),
                device=self.cdevice,
                dtype=self.dtype,
            )
        )
        nn.init.uniform_(self.initial_layer_weights, a=0.0, b=2 * math.pi)
        nn.init.uniform_(self.weights, a=0.0, b=2 * math.pi)

    def circuit(self, _: Optional[Tensor] = None) -> None:
        """
        Define the quantum circuit for this layer.

        :param _: Input tensor that is passed to the quantum circuit (ignored).
        :type _: Optional[Tensor]
        """
        for index, q in enumerate(self.wires):
            qml.Rot(
                self.initial_layer_weights[index, 0],
                self.initial_layer_weights[index, 1],
                self.initial_layer_weights[index, 2],
                q,
            )

        for layer in range(self.n_layers):
            for i in range(0, len(self.wires) - 1, 2):
                qml.CNOT(wires=[self.wires[i], self.wires[i + 1]])
                qml.Rot(
                    self.weights[layer, i, 0],
                    self.weights[layer, i, 1],
                    self.weights[layer, i, 2],
                    self.wires[i],
                )
                qml.Rot(
                    self.weights[layer, i + 1, 0],
                    self.weights[layer, i + 1, 1],
                    self.weights[layer, i + 1, 2],
                    self.wires[i + 1],
                )

            offset = int(self.num_wires / 2) * 2
            for i in range(1, len(self.wires) - 1, 2):
                qml.CNOT(wires=[self.wires[i], self.wires[i + 1]])
                qml.Rot(
                    self.weights[layer, offset + i - 1, 0],
                    self.weights[layer, offset + i - 1, 1],
                    self.weights[layer, offset + i - 1, 2],
                    self.wires[i],
                )
                qml.Rot(
                    self.weights[layer, offset + i, 0],
                    self.weights[layer, offset + i, 1],
                    self.weights[layer, offset + i, 2],
                    self.wires[i + 1],
                )


class AltRXCXLayer(CircuitLayer):
    """
    Layer for alternating CNOT gates with RX rotations.

    :param wires: The wires to be used by the layer.
    :type wires: Wires
    :param n_layers: The number of layers for the simplified two-design architecture,
        defaults to 1.
    :type n_layers: int, optional
    :param cdevice: Classical device to store the initial layer weights
        and internal layer weights, defaults to None.
    :type cdevice: CDevice, optional
    :param dtype: Data type of the weights, defaults to None.
    :type dtype: DType, optional
    :param kwargs: Extra arguments passed to the SimplifiedTwoDesign.
    """

    def __init__(
        self,
        wires: Wires,
        n_layers: int = 1,
        cdevice: Optional[CDevice] = None,
        dtype: Optional[DType] = None,
        **kwargs,
    ) -> None:
        super().__init__(wires)

        self.n_layers = n_layers
        self.cdevice = cdevice
        self.dtype = dtype
        self.kwargs = kwargs

        # weight parameters
        self.initial_layer_weights = torch.nn.Parameter(
            torch.empty((self.num_wires), device=self.cdevice, dtype=self.dtype)
        )
        self.weights = torch.nn.Parameter(
            torch.empty(
                (self.n_layers, 2 * (self.num_wires - 1)),
                device=self.cdevice,
                dtype=self.dtype,
            )
        )
        nn.init.uniform_(self.initial_layer_weights, a=0.0, b=2 * math.pi)
        nn.init.uniform_(self.weights, a=0.0, b=2 * math.pi)

    def circuit(self, _: Optional[Tensor] = None) -> None:
        """
        Define the quantum circuit for this layer.

        :param _: Input tensor that is passed to the quantum circuit (ignored).
        :type _: Optional[Tensor]
        """
        for index, q in enumerate(self.wires):
            qml.RX(
                self.initial_layer_weights[index],
                q,
            )

        for layer in range(self.n_layers):
            for i in range(0, len(self.wires) - 1, 2):
                qml.CNOT(wires=[self.wires[i], self.wires[i + 1]])
                qml.RX(
                    self.weights[layer, i],
                    self.wires[i],
                )
                qml.RX(
                    self.weights[layer, i + 1],
                    self.wires[i + 1],
                )

            offset = int(self.num_wires / 2) * 2
            for i in range(1, len(self.wires) - 1, 2):
                qml.CNOT(wires=[self.wires[i], self.wires[i + 1]])
                qml.RX(
                    self.weights[layer, offset + i - 1],
                    self.wires[i],
                )
                qml.RX(
                    self.weights[layer, offset + i],
                    self.wires[i + 1],
                )


class IQPERYCZLayer(CircuitLayer):
    """
    Layer combining an IQP embedding layer and an RY-CZ variational layer.

    :param wires: The wires to be used by the layer.
    :type wires: Wires
    :param num_uploads: The number of times to repeat data uploading, defaults to 1.
    :type num_uploads: int, optional
    :param num_varlayers: The number of times to repeat the variational layer, defaults to 1.
    :type num_varlayers: int, optional
    :param num_repeat: The number of times to repeat the combined layers, defaults to 1.
    :type num_repeat: int, optional
    :param base: The base of the exponent by which the inputs are scaled
        on each repetition, defaults to 1.0
    :type base: float, optional
    :param omega: The exponent for the base of the power by which the inputs are scaled
        on each repetition, defaults to 0.0
    :type omega: float, optional
    :param cdevice: Classical device to store the observable weights.
        If None specified, the default device is used.
    :type cdevice: CDevice, optional
    :param dtype: Data type of the variational weights.
    :type dtype: DType, optional
    :param iqpe_opts: Options for the IQPE class. Defaults to empty.
    :type iqpe_opts: Dict, optional
    :param rycz_opts: Options for the RYCZLayer class. Defaults to empty.
    :type rycz_opts: Dict, optional
    """

    def __init__(
        self,
        wires: Wires,
        num_uploads: int = 1,
        num_varlayers: int = 1,
        num_repeat: int = 1,
        base: Tensor = torch.tensor(1.0),
        omega: Tensor = torch.tensor(0.0),
        cdevice: Optional[CDevice] = None,
        dtype: Optional[DType] = None,
        iqpe_opts: Dict = {},
        rycz_opts: Dict = {},
    ) -> None:
        super().__init__(wires)
        self.num_uploads = num_uploads
        self.num_varlayers = num_varlayers
        self.num_repeat = num_repeat
        self.base = base
        self.omega = omega
        self.cdevice = cdevice
        self.dtype = dtype
        self.iqpe_opts = iqpe_opts
        self.rycz_opts = rycz_opts

        self.blocks = nn.ModuleList()

        num_var_repeats = 1
        if len(self.wires) == 1:
            num_var_repeats = self.num_varlayers

        for _ in range(self.num_repeat):
            embed_layer = IQPEmbeddingLayer(self.wires, self.num_uploads, **self.iqpe_opts)

            var_layers = []
            for _ in range(num_var_repeats):
                var_layer = RYCZLayer(
                    self.wires,
                    self.num_varlayers,
                    self.cdevice,
                    self.dtype,
                    **self.rycz_opts,
                )
                var_layers.append(var_layer)
            all_layers = [embed_layer] + var_layers
            block = nn.Sequential(*all_layers)
            self.blocks.append(block)

    def circuit(self, x: Tensor) -> None:
        """
        Define the quantum circuit for this layer.

        :param x: Input tensor that is passed to the quantum circuit.
        :type x: Tensor
        """
        for i, block in enumerate(self.blocks):
            fac = self.base ** (self.omega * i)
            block(fac * x)


class IQPEAltRotCXLayer(CircuitLayer):
    """
    Layer combining an IQP embedding layer and an alternating U3-CX variational layer.

    :param wires: The wires to be used by the layer.
    :type wires: Wires
    :param num_uploads: The number of times to repeat data uploading, defaults to 1.
    :type num_uploads: int, optional
    :param num_varlayers: The number of times to repeat the variational layer, defaults to 1.
    :type num_varlayers: int, optional
    :param num_repeat: The number of times to repeat the combined layers, defaults to 1.
    :type num_repeat: int, optional
    :param base: The base of the exponent by which the inputs are scaled
        on each repetition, defaults to 1.0
    :type base: float, optional
    :param omega: The exponent for the base of the power by which the inputs are scaled
        on each repetition, defaults to 0.0
    :type omega: float, optional
    :param cdevice: Classical device to store the observable weights.
        If None specified, the default device is used.
    :type cdevice: CDevice, optional
    :param dtype: Data type of the variational weights.
    :type dtype: DType, optional
    :param iqpe_opts: Options for the IQPE class. Defaults to empty.
    :type iqpe_opts: Dict, optional
    :param altrotcx_opts: Options for the AltRotCXLayer class. Defaults to empty.
    :type altrotcx_opts: Dict, optional
    """

    def __init__(
        self,
        wires: Wires,
        num_uploads: int = 1,
        num_varlayers: int = 1,
        num_repeat: int = 1,
        base: Tensor = torch.tensor(1.0),
        omega: Tensor = torch.tensor(0.0),
        cdevice: Optional[CDevice] = None,
        dtype: Optional[DType] = None,
        iqpe_opts: Dict = {},
        altrotcx_opts: Dict = {},
    ) -> None:
        super().__init__(wires)
        self.num_uploads = num_uploads
        self.num_varlayers = num_varlayers
        self.num_repeat = num_repeat
        self.base = base
        self.omega = omega
        self.cdevice = cdevice
        self.dtype = dtype
        self.iqpe_opts = iqpe_opts
        self.altrotcx_opts = altrotcx_opts

        self.blocks = nn.ModuleList()

        for _ in range(self.num_repeat):
            embed_layer = IQPEmbeddingLayer(self.wires, self.num_uploads, **self.iqpe_opts)
            var_layer = AltRotCXLayer(
                self.wires,
                self.num_varlayers,
                self.cdevice,
                self.dtype,
                **self.altrotcx_opts,
            )
            block = nn.Sequential(embed_layer, var_layer)
            self.blocks.append(block)

    def circuit(self, x: Tensor) -> None:
        """
        Define the quantum circuit for this layer.

        :param x: Input tensor that is passed to the quantum circuit.
        :type x: Tensor
        """
        for i, block in enumerate(self.blocks):
            fac = self.base ** (self.omega * i)
            block(fac * x)


class HadamardLayer(CircuitLayer):
    """
    A layer that adds Hadamard gates to each wire.

    :param wires: The wires to be used by the layer.
    :type wires: Wires
    """

    def __init__(self, wires: Wires):
        super().__init__(wires)
        self.qfunc = qml.Hadamard

    def circuit(self, _: Optional[Tensor] = None) -> None:
        """
        Apply the quantum circuit for this layer.

        :param _: Input tensor that is passed to the quantum circuit (ignored).
        :type _: Optional[Tensor]
        """
        for wire in self.wires:
            self.qfunc(wire)


class ParallelIQPEncoding(CircuitLayer):
    """
    A class that applies the IQPEmbedding to different parts of the input data.

    :param wires: The wires on which the circuit will be applied
    :type wires: Wires
    :param num_features: The number of features in the input
    :type num_features: int
    :param n_repeat: The number of times the IQPEmbedding will be repeated, defaults to 1
    :type n_repeat: int, optional
    :param base: The base of the exponent by which the inputs are scaled
        on each repetition, defaults to 1.0
    :type base: float, optional
    :param omega: The exponent for the base of the power by which the inputs are scaled
        on each repetition, defaults to 0.0
    :type omega: float, optional

    :raises ValueError: If the number of wires is less than the number of features
    :raises ValueError: If the number of wires is not a multiple of the number of features
    """

    def __init__(
        self,
        wires: Wires,
        num_features: int,
        n_repeat: int = 1,
        base: Tensor = torch.tensor(1.0),
        omega: Tensor = torch.tensor(0.0),
        **kwargs,
    ) -> None:
        super().__init__(wires)
        self.num_features = num_features
        self.n_repeat = n_repeat
        self.base = base
        self.omega = omega
        self.kwargs = kwargs
        self.qfunc = qml.IQPEmbedding

        if not self.num_wires >= self.num_features:
            raise ValueError(
                f"The number of wires ({self.num_wires}) "
                f"must be greater than or equal to the number of features ({self.num_features})."
            )
        if not self.num_wires % self.num_features == 0:
            raise ValueError(
                f"The number of wires ({self.num_wires}) "
                f"must be a multiple of the number of features ({self.num_features})."
            )

    def circuit(self, x: Tensor) -> None:
        """
        Apply the quantum circuit for this layer.

        :param x: Input tensor that is passed to the quantum circuit.
        :type x: Tensor
        """
        num_features = x.shape[-1]
        if num_features != self.num_features:
            raise ValueError(
                f"Input tensor last dimension ({num_features}) "
                f"must be equal to the number of features ({self.num_features})."
            )

        freq = 0
        for i in range(0, len(self.wires), num_features):
            x_ = self.base ** (freq * self.omega) * x
            self.qfunc(
                x_,
                self.wires[i : i + num_features],
                self.n_repeat,
                **self.kwargs,
            )
            freq += 1


class ParallelEntangledIQPEncoding(CircuitLayer):
    """
    A class that applies the IQPEmbedding on the entire constructed large feature vector.

    :param wires: The wires on which the circuit will be applied
    :type wires: Wires
    :param num_features: The number of features in the input
    :type num_features: int
    :param n_repeat: The number of times the IQPEmbedding will be repeated, defaults to 1
    :type n_repeat: int, optional
    :param base: The base of the exponent by which the inputs are scaled
        on each repetition, defaults to 1.0
    :type base: float, optional
    :param omega: The exponent for the base of the power by which the inputs are scaled
        on each repetition, defaults to 0.0
    :type omega: float, optional

    :raises ValueError: If the number of wires is less than the number of features
    :raises ValueError: If the number of wires is not a multiple of the number of features
    """

    def __init__(
        self,
        wires: Wires,
        num_features: int,
        n_repeat: int = 1,
        base: Tensor = torch.tensor(1.0),
        omega: Tensor = torch.tensor(0.0),
        **kwargs,
    ) -> None:
        super().__init__(wires)
        self.num_features = num_features
        self.n_repeat = n_repeat
        self.base = base
        self.omega = omega
        self.kwargs = kwargs
        self.qfunc = qml.IQPEmbedding

        if not self.num_wires >= self.num_features:
            raise ValueError(
                f"The number of wires ({self.num_wires}) "
                f"must be greater than or equal to the number of features ({self.num_features})."
            )
        if not self.num_wires % self.num_features == 0:
            raise ValueError(
                f"The number of wires ({self.num_wires}) "
                f"must be a multiple of the number of features ({self.num_features})."
            )

    def circuit(self, x: Tensor) -> None:
        """
        Apply the quantum circuit for this layer.

        :param x: Input tensor that is passed to the quantum circuit.
        :type x: Tensor
        """
        num_features = x.shape[-1]
        if num_features != self.num_features:
            raise ValueError(
                f"Input tensor last dimension ({num_features}) "
                f"must be equal to the number of features ({self.num_features})."
            )

        num_repeats = int(self.num_wires / num_features)

        x_large = []
        for j in range(0, num_repeats):
            x_ = self.base ** (j * self.omega) * x
            x_large.append(x_)

        x_final = torch.cat(x_large)
        self.qfunc(x_final, self.wires, self.n_repeat, **self.kwargs)


class TwoQubitRotCXMPSLayer(CircuitLayer):
    """
    Layer with 2-qubit MPS sctructure.

    :param wires: The wires to be used by the layer.
    :type wires: Wires
    :param n_layers_mps: The number of layers to repeat the MPS,
        defaults to 1.
    :type n_layers: int, optional
    :param n_layers_block: The number of layers for each block of the MPS,
        defaults to 1.
    :type n_layers_block: int, optional
    :param reverse: Flag to reverse the MPS sequence to bottom to top,
        defaults to False.
    :type reverse: bool, optional
    :param cdevice: Classical device to store the initial layer weights
        and internal layer weights, defaults to None.
    :type cdevice: CDevice, optional
    :param dtype: Data type of the weights, defaults to None.
    :type dtype: DType, optional
    """

    def __init__(
        self,
        wires: Wires,
        n_layers_mps: int = 1,
        n_layers_block: int = 1,
        reverse: bool = False,
        cdevice: Optional[CDevice] = None,
        dtype: Optional[DType] = None,
    ) -> None:
        super().__init__(wires)

        self.n_layers_mps = n_layers_mps
        self.n_layers_block = n_layers_block
        self.reverse = reverse
        self.cdevice = cdevice
        self.dtype = dtype
        self.n_blocks = self.num_wires - 1

        self.weights = torch.nn.Parameter(
            torch.empty(
                (self.n_layers_mps, self.n_blocks, 2, self.n_layers_block, 3),
                device=self.cdevice,
                dtype=self.dtype,
            )
        )
        self.weights_post = torch.nn.Parameter(
            torch.empty((self.num_wires, 3), device=self.cdevice, dtype=self.dtype)
        )

        nn.init.uniform_(self.weights, a=0.0, b=2 * math.pi)
        nn.init.uniform_(self.weights_post, a=0.0, b=2 * math.pi)

    def circuit(self, _: Optional[Tensor] = None) -> None:
        """
        Define the quantum circuit for this layer.

        :param _: Input tensor that is passed to the quantum circuit (ignored).
        :type x: Optional[Tensor]
        """

        for mps_layer_idx in range(self.n_layers_mps):
            for block_idx in (
                range(self.n_blocks - 1, -1, -1) if self.reverse else range(self.n_blocks)
            ):
                self._block(mps_layer_idx, block_idx)

        for i, q in enumerate(self.wires):
            qml.Rot(
                self.weights_post[i, 0],
                self.weights_post[i, 1],
                self.weights_post[i, 2],
                q,
            )

    def _block(self, mps_layer_idx, block_idx):
        qprev = self.wires[block_idx]
        qnext = self.wires[block_idx + 1]

        for block_layer in range(self.n_layers_block):
            qml.Rot(
                self.weights[mps_layer_idx, block_idx, 0, block_layer, 0],
                self.weights[mps_layer_idx, block_idx, 0, block_layer, 1],
                self.weights[mps_layer_idx, block_idx, 0, block_layer, 2],
                qprev,
            )
            qml.Rot(
                self.weights[mps_layer_idx, block_idx, 1, block_layer, 0],
                self.weights[mps_layer_idx, block_idx, 1, block_layer, 1],
                self.weights[mps_layer_idx, block_idx, 1, block_layer, 2],
                qnext,
            )
            qml.CNOT(wires=(qprev, qnext))


class EmbedU(CircuitLayer):
    """
    Layer that embeds an arbitrary unitary.

    :param wires: The wires to be used by the layer.
    :type wires: Wires
    :param U: The unitary to embed.
    :type U: Tensor.
    """

    def __init__(self, wires: Wires, U: Tensor) -> None:
        super().__init__(wires)
        self.U = U

    def circuit(self, _: Optional[Tensor] = None) -> None:
        """
        Define the quantum circuit for this layer.

        :param _: Input tensor that is passed to the quantum circuit (ignored).
        :type x: Optional[Tensor]
        """
        qml.QubitUnitary(self.U, wires=self.wires, unitary_check=False)


class MeasurementLayer(nn.Module):
    """
    Base class for measurment layers.

    Measurement layers are appended to quantum circuits and return classical output.

    :param circuits: Quantum circuits before the measurement.
    :type circuits: tuple
    :param qdevice: Quantum device. If None, the default device will be used.
    :type qdevice: Optional[QDevice], defaults to None
    :param measurement_type: Type of quantum measurement.
    :type measurement_type: MeasurementType, defaults to MeasurementType.Probabilities
    :param observables: Observables to measure.
        None only works with Probabilities and Samples.
    :type observables: Optional[Observables], defaults to None
    :param kwargs: Additional keyword arguments for qml.QNode.
    """

    def __init__(
        self,
        *circuits,
        qdevice: Optional[QDevice] = None,
        measurement_type: MeasurementType = MeasurementType.Probabilities,
        observables: Optional[Observables] = None,
        **kwargs,
    ) -> None:
        super().__init__()
        self.circuits = nn.ModuleList(circuits)

        if qdevice is None:
            self.qdevice = qml.device(wires=self.circuits[0].wires, **DEFAULT_QDEV_CFG)
        else:
            self.qdevice = qdevice

        self.wires = self.circuits[0].wires
        self.measurement_type = measurement_type

        self.observables = observables
        if observables is not None and not isinstance(observables, Iterable):
            self.observables = [observables]

        self.interface = kwargs.pop("interface", "torch")
        self.diff_method = kwargs.pop("diff_method", "backprop")
        self.kwargs = kwargs
        self.check_measurement_type()
        self.qnode = self.set_qnode()

        self.subwires = [0]

    def forward(self, x: Optional[Tensor] = None) -> Tensor:
        """
        Forward pass, depending on measurement type.
        See :meth:`expectation`, :meth:`probabilities` and :meth:`samples`.

        :param x: Input data samples.
        :type circuits: Tensor.
        :return: Forward evaluation of model on data.
        :rtype: Tensor
        """
        self.qnode = self.set_qnode()

        if x is not None:
            if len(x.shape) == 1:
                out = self.qnode(x)
            else:
                outs = [self.qnode(xk) for xk in torch.unbind(x)]
                out = torch.stack(outs)

            if (len(x.shape) == 1 and len(out.shape) == 0) or (
                len(x.shape) > 1 and len(out.shape) == 1
            ):
                out = out.unsqueeze(-1)

        else:
            out = self.qnode(x)

        return out

    def expectation(self, x: Optional[Tensor] = None) -> Expectation:
        """
        Calculate the expectation value of the observable for given circuits.

        :param x: Input tensor that is passed to the quantum circuits, defaults to None.
        :type x: Optional[Tensor]
        :return: Expectation value object for the observable.
        :rtype: Expectation
        """
        for circuit in self.circuits:
            circuit(x)

        expec = None
        if self.observables is not None:
            expec = [qml.expval(obs) for obs in self.observables]

        return expec

    def probabilities(self, x: Optional[Tensor] = None) -> Probability:
        """
        Calculate the outcome probabilities for given circuits.

        :param x: Input tensor that is passed to the quantum circuits, defaults to None.
        :type x: Optional[Tensor]
        :return: Probabilities of the outcomes of the circuits.
        :rtype: Probability
        """
        for circuit in self.circuits:
            circuit(x)
        probs = qml.probs(wires=self.wires)
        return probs

    def samples(self, x: Optional[Tensor] = None) -> Sample:
        """
        Sample the outcomes of the given circuits.

        :param x: Input tensor that is passed to the quantum circuits, defaults to None.
        :type x: Optional[Tensor]
        :return: Samples of the outcomes of the circuits.
        :rtype: Sample
        """
        for circuit in self.circuits:
            circuit(x)
        sample = qml.sample(wires=self.wires)
        return sample

    def entropy(self, x: Optional[Tensor] = None) -> Entropy:
        """
        Calculate the Von Neumann Entropy of a subsystem.

        :param x: Input tensor that is passed to the quantum circuits, defaults to None.
        :type x: Optional[Tensor]
        :return: Entropy value object for a subsystem.
        :rtype: Entropy
        """
        for circuit in self.circuits:
            circuit(x)
        entropy = qml.vn_entropy(self.subwires)
        return entropy

    def set_qnode(self) -> QNode:
        """
        Set the quantum node for the layer and measurement type.

        :return: The set QNode.
        :rtype: QNode
        """

        if self.measurement_type == MeasurementType.Expectation:
            circuit = self.expectation
        elif self.measurement_type == MeasurementType.Probabilities:
            circuit = self.probabilities
        elif self.measurement_type == MeasurementType.Samples:
            circuit = self.samples
        elif self.measurement_type == MeasurementType.Entropy:
            circuit = self.entropy

        qnode = qml.QNode(
            circuit,
            self.qdevice,
            interface=self.interface,
            diff_method=self.diff_method,
            **self.kwargs,
        )
        self.qnode = qnode
        return self.qnode

    def check_measurement_type(self) -> None:
        """
        Check if the measurement type is valid.
        Raises errors for invalid measurement types.

        :raises NotImplementedError: If the measurement type is not recognized.
        :raises ValueError: If the expectation measurement type doesn't have an
            observable, or if the sample measurement type doesn't have an
            integer number of shots.
        """

        if not isinstance(self.measurement_type, MeasurementType):
            raise NotImplementedError(f"Measurement type ({self.measurement_type}) not recognized")
        if self.measurement_type == MeasurementType.Expectation:
            if self.observables is None:
                raise ValueError(
                    f"Measurement type ({self.measurement_type}) " "requires an observable"
                )
        if self.measurement_type == MeasurementType.Samples and self.qdevice.shots is None:
            raise ValueError(
                f"Measurement type ({self.measurement_type}) " "requires integer number of shots"
            )


class HamiltonianLayer(MeasurementLayer):
    """
    A layer that computes the expectation of a Hamiltonian.

    The Hamiltonian is defined by a list of observables and their associated weights.
    The weights are trainable parameters.

    :param circuits: Quantum circuits that make up the circuit layer before measurement.
    :type circuits: tuple
    :param observables: Observables defining the Hamiltonian.
    :type observables: list of Observable
    :param qdevice: Quantum device. If None specified, the default device is used.
    :type qdevice: Optional[QDevice]
    :param cdevice: Classical device to store the observable weights.
        If None specified, the default device is used.
    :type cdevice: CDevice, optional
    :param dtype: Data type of the observable weights.
    :type dtype: DType, optional
    :param kwargs: Additional keyword arguments passed to the superclass.
    """

    def __init__(
        self,
        *circuits,
        observables: Iterable[Observable],
        qdevice: Optional[QDevice] = None,
        cdevice=None,
        dtype=None,
        **kwargs,
    ) -> None:
        super().__init__(
            *circuits,
            qdevice=qdevice,
            measurement_type=MeasurementType.Expectation,
            observables=observables,
            **kwargs,
        )
        self.cdevice = cdevice
        self.dtype = dtype

        # set observable weights
        self.num_weights = len(list(observables))
        self.observable_weights = torch.nn.Parameter(
            torch.empty(self.num_weights, device=self.cdevice, dtype=self.dtype)
        )
        nn.init.normal_(self.observable_weights)
        self.observable = qml.Hamiltonian(self.observable_weights, observables)

    def expectation(self, x: Optional[Tensor] = None) -> Expectation:
        """
        Compute the expectation of the Hamiltonian.

        :param x: Input tensor, defaults to None.
        :type x: Optional[Tensor]
        :return: Expectation value of the Hamiltonian.
        :rtype: Expectation
        """
        for circuit in self.circuits:
            circuit(x)
        self.observable = qml.Hamiltonian(self.observable_weights, self.observables)
        expec = qml.expval(self.observable)
        return expec



from enum import Enum
from typing import Dict, Optional

import torch
from torch import nn

from .qkernel import QKernel
from .types import DataLoader, Logger, Loss, Metric, Model, Optimizer, Parameter, Tensor, Writer


class EpochType(Enum):
    """
    Enum to denote the type of epoch in the training process.

    :cvar Train: Indicates the epoch is a training epoch.
    :cvar Validate: Indicates the epoch is a validation epoch.
    """

    Train = 1
    Validate = 2


class SupervisedTrainer:
    """
    Class to handle the training of a supervised learning model.

    :param optimizer: The optimizer to be used in the training process.
    :type optimizer: Optimizer
    :param loss_fn: The loss function used for optimization.
    :type loss_fn: Loss
    :param num_epochs: The number of epochs to train for.
    :type num_epochs: int
    :param metrics: A dictionary mapping metric names to the metric functions to be evaluated.
        Defaults to empty dict.
    :type metrics: Dict[str, Metric], optional
    :param writer: An optional writer for logging purposes. Default is None.
    :type writer: Optional[Writer]
    :param logger: An optional logger for logging purposes. Default is None.
    :type logger: Optional[Logger]
    """

    def __init__(
        self,
        optimizer: Optimizer,
        loss_fn: Loss,
        num_epochs: int,
        metrics: Dict[str, Metric] = {},
        writer: Optional[Writer] = None,
        logger: Optional[Logger] = None,
    ) -> None:
        self.optimizer = optimizer
        self.loss_fn = loss_fn
        self.metrics = metrics
        self.num_epochs = num_epochs
        self.writer = writer
        self.logger = logger

    def train(self, model: Model, train_data: DataLoader, valid_data: DataLoader) -> None:
        """
        Train the given model using the provided data loaders.

        :param model: The model to be trained.
        :type model: Model
        :param train_data: The DataLoader for the training data.
        :type train_data: DataLoader
        :param valid_data: The DataLoader for the validation data.
        :type valid_data: DataLoader
        """

        for epoch in range(1, self.num_epochs + 1):
            self.train_epoch(model, train_data, epoch)
            self.validate_epoch(model, valid_data, epoch)

    def train_epoch(self, model: Model, train_data: DataLoader, epoch: int = 0) -> None:
        """
        Train the model for one epoch.

        :param model: The model to be trained.
        :type model: Model
        :param train_data: The DataLoader for the training data.
        :type train_data: DataLoader
        :param epoch: The current epoch number. Default is 0.
        :type epoch: int
        """
        epoch_type = EpochType.Train
        self._epoch(epoch_type, model, train_data, epoch)

    def validate_epoch(self, model: Model, valid_data: DataLoader, epoch: int = 0) -> None:
        """
        Validate the model after an epoch of training.

        :param model: The model to be validated.
        :type model: Model
        :param valid_data: The DataLoader for the validation data.
        :type valid_data: DataLoader
        :param epoch: The current epoch number. Default is 0.
        :type epoch: int
        """
        epoch_type = EpochType.Validate
        self._epoch(epoch_type, model, valid_data, epoch)

    def _epoch(self, epoch_type: EpochType, model: Model, data: DataLoader, epoch: int = 0) -> None:
        running_loss = 0.0
        running_metrics = {}
        for metric in self.metrics:
            running_metrics[metric] = 0.0

        if epoch_type == EpochType.Train:
            for inputs, labels in data:
                self._train_step(model, inputs, labels)

        for inputs, labels in data:
            with torch.no_grad():
                predicted = model(inputs)
                loss = self.loss_fn(predicted, labels)
                running_loss += loss.item() * len(inputs)
                for metric in self.metrics:
                    metric_val = self.metrics[metric](predicted, labels)
                    running_metrics[metric] += metric_val.item() * len(inputs)

        running_loss /= float(len(data.dataset))  # type: ignore
        for metric in self.metrics:
            running_metrics[metric] /= float(len(data.dataset))  # type: ignore

        phase = epoch_type.name
        self._log_metrics(phase, running_loss, running_metrics, epoch)

    def _train_step(self, model: Model, inputs: Tensor, labels: Tensor) -> None:
        self.optimizer.zero_grad()
        predicted = model(inputs)
        loss = self.loss_fn(predicted, labels)
        loss.backward()
        self.optimizer.step()

    def _log_metrics(self, phase: str, loss: float, metrics: Dict[str, float], epoch: int) -> None:
        if self.writer is not None:
            self.writer.add_scalar(f"Loss/{phase}", loss, epoch)
            for metric_name, metric_value in metrics.items():
                self.writer.add_scalar(f"Metrics/{phase}/{metric_name}", metric_value, epoch)

        if self.logger is not None:
            metrics_strs = [
                f"{metric_name}: {metric_value:.6f}"
                for metric_name, metric_value in metrics.items()
            ]
            self.logger.info(
                f"{phase} - Epoch: {epoch}, Loss: {loss:.6f}, Metrics: {', '.join(metrics_strs)}"
            )


class RidgeRegression:
    """
    Class to handle the training of a model using Ridge Regression.

    :param lambda_reg: Regularization strength. Must be positive.
    :type lambda_reg: float
    :param metrics: A dictionary mapping metric names to the metric functions to be evaluated.
        Defaults to empty dict.
    :type metrics: Dict[str, Metric], optional
    :param logger: An optional logger for logging purposes. Default is None.
    :type logger: Optional[Logger]
    """

    def __init__(
        self,
        lambda_reg: float,
        metrics: Dict[str, Metric] = {},
        logger: Optional[Logger] = None,
    ) -> None:
        self.lambda_reg = lambda_reg
        self.metrics = metrics
        self.logger = logger

    def train(self, model: QKernel, train_data: DataLoader, valid_data: DataLoader) -> None:
        """
        Train the given model using the provided data loaders using Ridge Regression.

        :param model: The quantum kernel model to be trained.
        :type model: QKernel
        :param train_data: The DataLoader for the training data.
        :type train_data: DataLoader
        :param valid_data: The DataLoader for the validation data.
        :type valid_data: DataLoader

        .. warning::
            Training changes the state of the model by assigning `X_train`.
        """
        if len(train_data) != 1:
            raise ValueError("For ridge regression batching training data is invalid")
        if len(valid_data) != 1:
            raise ValueError("For ridge regression batching validation data is invalid")

        for inputs, labels in train_data:
            model.X_train = inputs
            alpha = self.kernel_ridge_regression(model, inputs, labels)
            model.alpha = alpha

        phase = EpochType.Train.name
        running_metrics = {}
        for inputs, labels in train_data:
            predicted = model(inputs)
            for metric in self.metrics:
                running_metrics[metric] = self.metrics[metric](predicted, labels)
        self._log_metrics(phase, running_metrics)

        phase = EpochType.Validate.name
        running_metrics = {}
        for inputs, labels in valid_data:
            predicted = model(inputs)
            for metric in self.metrics:
                running_metrics[metric] = self.metrics[metric](predicted, labels)
        self._log_metrics(phase, running_metrics)

    def kernel_ridge_regression(self, model: QKernel, inputs: Tensor, labels: Tensor) -> Parameter:
        """
        Compute Ridge Regression solution for the given inputs and labels using the provided model.

        :param model: The quantum kernel model.
        :type model: QKernel
        :param inputs: Input data tensor.
        :type inputs: Tensor
        :param labels: Corresponding labels tensor.
        :type labels: Tensor
        :return: The computed alpha parameter tensor.
        :rtype: Parameter
        """
        linputs = len(inputs.shape)
        if linputs != 2:
            raise ValueError(f"Inputs must have 2 dimensions each, not {linputs}")

        K = model.kernel_matrix(inputs, inputs)
        num_samples = inputs.shape[0]
        Id = torch.eye(num_samples, dtype=labels.dtype, device=labels.device)
        M = K + self.lambda_reg * Id
        alpha = nn.Parameter(torch.linalg.solve(M, labels))

        return alpha

    def _log_metrics(self, phase: str, metrics: Dict[str, float]) -> None:
        """
        Log computed metrics for the provided phase.

        :param phase: The phase (train or validate) being logged.
        :type phase: str
        :param metrics: Dictionary of computed metrics.
        :type metrics: Dict[str, float]
        """
        if self.logger is not None:
            metrics_strs = [
                f"{metric_name}: {metric_value:.6f}"
                for metric_name, metric_value in metrics.items()
            ]
            self.logger.info(f"{phase} - Metrics: {', '.join(metrics_strs)}")



import torch

from .datagen import DataGenRademacher as Datagen
from .loss import RademacherLoss
from .trainer import SupervisedTrainer as Trainer
from .types import QModel, Tensor


def rademacher(
    model: QModel, trainer: Trainer, X: Tensor, sigmas: Tensor, datagen: Datagen
) -> Tensor:
    """
    Estimate Rademacher complexity of a given model.

    :param model: Prediction model.
    :type model: QModel
    :param trainer: The trainer.
    :type trainer: Trainer
    :param X: Data tensor of size (num_data_samples, size_data_set, dim_feature)
    :type X: Tensor
    :param sigmas: Sigmas tensor of size (num_sigma_samples, size_data_set)
    :type sigmas: Tensor
    :param datagen: Datagen object for converting to loader.
    :type datagen: Datagen
    :return: Scalar-valued tensor with Rademacher complexity.
    :rtype: Tensor
    """

    num_data_samples = X.shape[0]
    num_sigma_samples = sigmas.shape[0]
    data = {"X": X, "sigmas": sigmas}

    sum = torch.tensor([0.0], requires_grad=False, device=X.device)
    for m in range(num_data_samples):
        for s in range(num_sigma_samples):
            X_ = X[m]
            sigmas_ = sigmas[s]
            loss_fn = RademacherLoss(sigmas_)
            trainer.loss_fn = loss_fn
            loader = datagen.data_to_loader(data, m)

            trainer.train(model, loader, loader)
            with torch.no_grad():
                predictions = model(X_)
                sum += -loss_fn(predictions)

    sum /= num_data_samples * num_sigma_samples

    return sum



import os

import toml


def get_version():
    here = os.path.dirname(os.path.realpath(__file__))
    version_info = toml.load(os.path.join(here, "../pyproject.toml"))
    return version_info["tool"]["poetry"]["version"]


__version__ = get_version()



"""Frequently used functions."""

import math
from itertools import chain, combinations
from typing import Dict, List, Tuple

from .types import Observable, Tensor


def probabilities_to_dictionary(probs: Tensor) -> Dict[str, Tensor]:
    """
    Convert 1D tensor of probabilities to dictionary of probabilities of bitstrings.

    :param probs: Probabilities.
    :type probs: Tensor
    :return: Probabilities of bitstrings.
    :rtype: Dict
    :raises ValueError: If length of probs is not a power of 2.
    """

    n = int(math.log2(len(probs)))
    if 2**n != len(probs):
        raise ValueError(f"Length of probs ({len(probs)}) is not a power of 2.")

    result = {}
    for i, p in enumerate(probs):
        bistring = bin(i)[2:].zfill(n)
        result[bistring] = p

    return result


def samples_to_dictionary(samples: Tensor) -> Dict[str, float]:
    """Convert 2D tensor of samples to dictionary of probabilities of bitstrings.

    :param samples: Samples.
    :type samples: Tensor
    :return: Probabilities of bitstrings.
    :rtype: Dict
    :raises ValueError: If samples not a Tensor of integers.
    """

    if samples.is_floating_point() or samples.is_complex():
        raise ValueError("Samples must be tensors of integers.")

    bitstrings = ["".join(str(b.item()) for b in sample) for sample in samples]
    bitstring_counts = {bs: bitstrings.count(bs) for bs in set(bitstrings)}
    total = sum(bitstring_counts.values())
    bitstring_probs = {bs: count / total for bs, count in bitstring_counts.items()}

    return bitstring_probs


def all_bin_sequences(n: int) -> List[Tuple[int, ...]]:
    """
    Generates all possible binary sequences of length n.

    :param n: The length of the binary sequences.
    :type n: int
    :return: A list of all binary sequences of length n,
        represented as a list of integers.
    :rtype: List[List[int]]
    """

    elements = list(range(n))
    return list(chain.from_iterable(combinations(elements, r) for r in range(n + 1)))


def parities_outcome(bitstring: str, H: Observable) -> float:
    """
    Compute the measurement outcome for a given bit string and Hamiltonian.
    Only works for Hamiltonians with identity and Pauli Z.

    :param bitstring: Input bit string.
    :type bitstring: string: str
    :param H: Hamiltonian.
    :type H: Observable
    :return: Real-valued outcome.
    :rtype: float
    :raises ValueError: If number of qubits does not match or
        operators other than I or Z are detected.
    """

    num_qubits = len(bitstring)
    num_wires = len(H.wires)

    if num_qubits != num_wires:
        raise ValueError(
            f"Number of qubits ({num_qubits}) " f"does not match number of wires ({num_wires})"
        )

    sum = 0.0
    for idx, O in enumerate(H.ops):
        if not isinstance(O.name, list):
            if O.name == "Identity":
                sum += H.coeffs[idx]
            elif O.name == "PauliZ":
                i = O.wires[0]
                sign = (-1) ** (int(bitstring[-1 - i]))
                sum += sign * H.coeffs[idx]
            else:
                raise ValueError("All operators must be PauliZ or Identity.")

        else:
            if not all(name == "PauliZ" for name in O.name):
                raise ValueError("All operators must be PauliZ or Identity.")

            sign = 1
            for w in O.wires:
                sign *= (-1) ** (int(bitstring[-1 - w]))

            sum += sign * H.coeffs[idx]

    return sum


def parities_outcome_probs(probs: Dict[str, float], H: Observable) -> Dict[float, float]:
    """
    Compute (real-valued) outputs with corresponding probabilities.

    :param probs: Dictionary of probabilities of bitstrings.
    :type probs: Dict
    :param H: Hamiltonian determening the outcome.
    :type H: Observable
    :return: Values with corresponding probabilities.
    :rtype: Dict
    """

    result = {}
    for b, p in probs.items():
        val = parities_outcome(b, H)
        result[val] = p

    return result



import logging

import numpy as np
import torch

from .datagen import DataGenCapacity as Datagen
from .trainer import SupervisedTrainer as Trainer
from .types import Capacity, Model


def memory(
    model: Model,
    datagen: Datagen,
    trainer: Trainer,
    Nmin: int,
    Nmax: int,
    Nstep: int = 1,
    early_stop: bool = True,
    stop_count: int = 1,
) -> Capacity:
    """
    Estimates the memory capacity of a model over a range of values of N.

    :param model: The model.
    :type model: Model
    :param datagen: The (synthetic) data generator.
    :type datagen: Datagen
    :param trainer: The trainer.
    :type trainer: Trainer
    :param Nmin: The minimum value of N.
    :type Nmin: int
    :param Nmax: The maximum value of N, included.
    :type Nmax: int
    :param Nstep: Step size for N. Defaults to 1.
    :type Nstep: int, optional
    :param early_stop: Stops early if previous stop_count iterations
            capacity at least as large. Defaults to True.
    :type early_stop: bool, optional
    :param stop_count: See early_stop. Defaults to 1.
    :type stop_count: int, optional
    :return: List of tuples (N, mre=2^(-m), m, N*m).
    :rtype: Capacity

    .. seealso::
        arXiv:1908.01364.
    """

    capacities = []
    Cprev = 0
    count = 0
    for N in range(Nmin, Nmax + 1, Nstep):
        mre = fit_rand_labels(model, datagen, trainer, N)
        m = max(int(np.log2(1.0 / mre)), 0)
        C = N * m

        capacities.append((N, mre, m, C))

        if C <= Cprev and N != Nmax and early_stop:
            count += 1
            if count >= stop_count:
                logging.basicConfig(level=logging.WARNING)
                logging.warning("Stopping early, capacity not improving.")
                break
        else:
            count = 0

        Cprev = C

    return capacities


def fit_rand_labels(model: Model, datagen: Datagen, trainer: Trainer, N: int) -> float:
    """
    Fits random labels to a model and returns the mean relative error.

    :param model: The model.
    :type model: Model
    :param datagen: The data generation class.
    :type datagen: Datagen
    :param trainer: The trainer.
    :type trainer: Trainer
    :param N: The number of inputs.
    :type N: int
    :return: The mean relative error.
    :rtype: float
    """

    data = datagen.gen_data(N)
    X = data["X"]
    Y = data["Y"]
    mre_sample = []
    for s in range(datagen.num_samples):
        loader = datagen.data_to_loader(data, s)
        trainer.train(model, loader, loader)
        y_pred = model(X)
        mre = torch.mean(torch.abs((Y[s] - y_pred) / y_pred))
        mre_sample.append(mre.item())

    mre_N = torch.mean(torch.tensor(mre_sample))
    return mre_N.item()



from typing import Any, Callable, Dict, Iterable, List, Sequence, Tuple, Union

# for python < 3.10
try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

import logging

import numpy as np
import pennylane as qml
import tntorch
import torch
from torch.utils.tensorboard import SummaryWriter

# Type aliases
Tensor: TypeAlias = torch.Tensor
Array: TypeAlias = np.ndarray
Device: TypeAlias = torch.device
DataOut: TypeAlias = Dict[str, Tensor]
DataLoader: TypeAlias = torch.utils.data.DataLoader
TensorDataset: TypeAlias = torch.utils.data.TensorDataset
Model: TypeAlias = torch.nn.Module
QModel: TypeAlias = qml.QNode
ParameterList: TypeAlias = List[List[Tensor]]
CDevice: TypeAlias = torch.device
DType: TypeAlias = torch.dtype
Loss: TypeAlias = torch.nn.Module
Capacity = List[Tuple[int, float, int, int]]
MPS: TypeAlias = tntorch.tensor.Tensor
Observable: TypeAlias = qml.operation.Observable
Observables: TypeAlias = Union[qml.operation.Observable, Iterable[qml.operation.Observable]]
Probability: TypeAlias = qml.measurements.ProbabilityMP
Sample: TypeAlias = qml.measurements.SampleMP
Entropy: TypeAlias = qml.measurements.VnEntropyMP
Hamiltonian: TypeAlias = qml.Hamiltonian
ParitySequence: TypeAlias = Sequence[Tuple[int, ...]]
QDevice: TypeAlias = qml.Device
QNode: TypeAlias = qml.QNode
Expectation: TypeAlias = qml.measurements.ExpectationMP
Wires: TypeAlias = Union[int, Iterable[Any]]
Optimizer: TypeAlias = torch.optim.Optimizer
Metric: TypeAlias = Callable
Writer: TypeAlias = SummaryWriter
Logger: TypeAlias = logging.Logger
Parameter: TypeAlias = torch.nn.Parameter



from typing import Optional

import pennylane as qml
import torch
from torch import nn

from .qlayer import CircuitLayer as FeatureEmbed

DEFAULT_QDEV_CFG = {"name": "default.qubit", "shots": None}

from .types import Expectation, QDevice, QNode, Tensor


class QKernel(nn.Module):
    """
    A quantum kernel module that represents a quantum embedding kernel.

    :param embed: The quantum feature embedding mechanism.
    :type embed: FeatureEmbed
    :param X_train: Training data tensor.
    :type X_train: Tensor
    :param qdevice: Quantum device. If None, the default device will be used.
    :type qdevice: Optional[QDevice], defaults to None
    :param kwargs: Additional keyword arguments for qml.QNode.
    """

    def __init__(
        self,
        embed: FeatureEmbed,
        X_train: Tensor,
        qdevice: Optional[QDevice] = None,
        **kwargs,
    ) -> None:
        super().__init__()
        self.embed = embed
        self._X_train = X_train
        self.num_samples = X_train.shape[0]
        self.alpha = nn.Parameter(
            torch.empty(
                self.num_samples,
                device=self.X_train.device,
                dtype=self.X_train.dtype,
            )
        )
        nn.init.normal_(self.alpha)

        if qdevice is None:
            self.qdevice = qml.device(wires=self.embed.wires, **DEFAULT_QDEV_CFG)
        else:
            self.qdevice = qdevice

        self.interface = kwargs.pop("interface", "torch")
        self.diff_method = kwargs.pop("diff_method", "backprop")
        self.qnode = self.set_qnode()

    @property
    def X_train(self) -> Tensor:
        return self._X_train

    @X_train.setter
    def X_train(self, X: Tensor) -> None:
        self._X_train = X
        self.num_samples = X.shape[0]
        self.alpha = nn.Parameter(
            torch.empty(
                self.num_samples,
                device=self.X_train.device,
                dtype=self.X_train.dtype,
            )
        )
        nn.init.normal_(self.alpha)

    def kernel_circuit(self, x: Tensor, x_: Tensor) -> Expectation:
        """
        Define the quantum circuit that calculates the kernel between input vectors.

        :param x: Input tensor representing the first data point.
        :type x: Tensor
        :param x_: Input tensor representing the second data point.
        :type x_: Tensor

        :return: Expectation value of the projected state.
        :rtype: Expectation
        """

        self.embed(x)
        qml.adjoint(self.embed)(x_)
        state = torch.zeros(self.embed.num_wires, dtype=torch.int32, device=x.device)
        projector = qml.Projector(state, self.embed.wires)

        return qml.expval(projector)

    def kernel_matrix(self, x: Tensor, x_: Tensor) -> Tensor:
        """
        Compute the kernel matrix between two sets of input vectors.

        :param x: Input tensor representing the first set of data points.
        :type x: Tensor
        :param x_: Input tensor representing the second set of data points.
        :type x_: Tensor

        :return: Kernel matrix of shape (number of data points in x, number of data points in x_).
        :rtype: Tensor

        :raises ValueError: If the input tensors do not have exactly 2 dimensions.
        """

        lx = len(x.shape)
        lx_ = len(x_.shape)
        if lx != 2 or lx_ != 2:
            raise ValueError(f"Inputs must have 2 dimensions each, not {lx} and {lx_}")

        K = torch.empty((x.shape[0], x_.shape[0]), dtype=x.dtype, device=x.device)
        for i, xi in enumerate(x):
            for j, xj in enumerate(x_):
                Kij = self.qnode(xi, xj)
                K[i, j] = Kij

        return K

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward computation of the quantum kernel using the training data and alpha parameters.

        :param x: Input tensor representing the data points.
        :type x: Tensor

        :return: Resultant tensor after the forward pass.
        :rtype: Tensor
        """

        K = self.kernel_matrix(x, self.X_train)
        out = torch.matmul(K, self.alpha)
        return out

    def set_qnode(self) -> QNode:
        """
        Define and set the quantum node for the kernel circuit.

        :return: The defined quantum node.
        :rtype: QNode
        """

        circuit = self.kernel_circuit
        qnode = qml.QNode(
            circuit,
            self.qdevice,
            interface=self.interface,
            diff_method=self.diff_method,
        )
        self.qnode = qnode
        return self.qnode



import math
from math import pi
from typing import List

import torch

from .types import Model, ParameterList, Tensor


def compute_effdim(
    model: Model,
    features: Tensor,
    param_list: ParameterList,
    weights: Tensor,
    volume: Tensor,
    gamma: Tensor,
) -> Tensor:
    """
    Compute the effective dimension as proposed in arXiv:2112.04807.

    :param model: The statistical model, which outputs a probability vector.
    :type model: Model
    :param features: The input data (X) used to compute the FIMs, shape (num_samples, num_features).
    :type features: Tensor
    :param param_list: A list of lists of parameters. Each set of parameters is used to
        compute an empirical FIM.
    :param weights: Weights for the Monte Carlo integration. The integral f(theta) is estimated
        as 1/N * sum_k f_k w_k.
    :type weights: Tensor
    :param volume: Volume of the parameter space.
    :type volume: Tensor
    :param gamma: Gamma parameter of the effective dimension.
    :type gamma: Tensor

    :return: Effective dimension.
    :rtype: Tensor

    .. note::
        The effective dimension will be negative for data samples that are too small,
        where n/log(n) < 2pi/gamma.
    """

    num_data_samples = len(features)
    num_parameters = sum(p.numel() for p in model.parameters() if p.requires_grad)

    fims = compute_fims(model, features, param_list)
    trace_integral = mc_integrate_fim_trace(fims, weights)
    norm_const = norm_const_fim(trace_integral, num_parameters, volume)
    kappa = const_effdim(num_data_samples, gamma)
    effdim = mc_integrate_fims_effdim(fims, weights, norm_const, kappa, volume)

    return effdim


def mc_integrate_fims_effdim(
    fims: List[Tensor],
    weights: Tensor,
    norm_const: Tensor,
    kappa: Tensor,
    volume: Tensor,
) -> Tensor:
    """
    Performs a weighted Monte Carlo integration of the square root of the determinant
    of the matrix (Identity + c*FIM(theta)) and estimate the effective dimension
    as in arXiv:2112.04807.

    :param fims: List of Fisher Information Matrices estimated at different parameter samples.
    :type fims: List[Tensor]
    :param weights: Weights for each parameter sample, or a single constant weight.
        The integral f(theta) is estimated as 1/N * sum_k f_k w_k.
    :type weights: Tensor
    :param norm_const: The FIM normalization constant.
    :type norm_const: Tensor
    :param kappa: The denominator in effective dimension.
    :type kappa: Tensor
    :param volume: Volume of the parameter space.
    :type volume: Tensor

    :return: A scalar tensor of the estimated effective dimension.
    :rtype: Tensor

    :raises ValueError: If any of the FIMs is not a square matrix, or if the number
        of FIMs is not equal to the number of weights (for non-scalar weights).
    """

    num_samples = len(fims)
    if weights.numel() > 1:
        num_weights = len(weights)
        if num_samples != num_weights:
            raise ValueError(
                f"Number of samples ({num_samples}) "
                "must be equal to number of weights ({num_weights})"
            )

    logdets = torch.zeros(num_samples, device=fims[0].device, dtype=fims[0].dtype)

    for i, fim in enumerate(fims):
        _check_fim(fim)
        logdet = half_log_det(fim, norm_const * kappa)
        logdets[i] = logdet
    zeta = torch.max(logdets)

    sum = torch.zeros(1, device=fims[0].device, dtype=fims[0].dtype)
    for logdet in logdets:
        if weights.numel() > 1:
            weight = weights[i]
        else:
            weight = weights

        sum += weight * torch.exp(logdet - zeta)
    sum /= num_samples

    result = 2.0 * zeta / torch.log(kappa) + 2.0 / torch.log(kappa) * torch.log(1.0 / volume * sum)

    return result


def half_log_det(fim: Tensor, c: Tensor) -> Tensor:
    """
    Estimates half of the logarithm of the determinant of the matrix (Identity + c*FIM).

    This function takes as input a square Fisher Information Matrix (FIM) and a scaling factor,
    and returns half of the logarithm of the determinant of the matrix (Identity + c*FIM).

    :param fim: The square Fisher Information Matrix (FIM).
    :type fim: Tensor
    :param c: The scaling factor.
    :type c: Tensor

    :return: Half of the logarithm of the determinant of the matrix (Identity + c*FIM).
    :rtype: Tensor

    :raises ValueError: If the FIM is not a square matrix.
    """

    _check_fim(fim)

    eigs = torch.linalg.eigvalsh(fim)
    result = torch.tensor(0.5, device=fim.device, dtype=fim.dtype) * sum(torch.log(1.0 + c * eigs))
    return result


def const_effdim(num_samples: int, gamma: Tensor) -> Tensor:
    """
    Computes the constant factor in the effective dimension as per the formula described in
    arXiv:2011.00027.

    This function takes the number of data samples and the effective dimension gamma parameter
    (between 0 and 1), and returns the constant factor in the effective dimension formula.

    :param num_samples: Number of data samples.
    :type num_samples: int
    :param gamma: The effective dimension gamma parameter (should be between 0 and 1, exclusive).
    :type gamma: float

    :return: The constant factor in the effective dimension formula.
    :rtype: float

    :raises ValueError: If the number of samples is less than 2 or the gamma parameter is not in
        the (0, 1] range.
    """

    if num_samples < 2:
        raise ValueError(f"num_samples ({num_samples}) has to be at least 2")
    if gamma <= 0.0 or gamma > 1.0:
        raise ValueError(f"gamma ({gamma}) has to be in (0, 1])")

    const = gamma * num_samples / (2 * pi * math.log(num_samples))
    return const


def norm_const_fim(trace_integral: Tensor, num_parameters: int, volume: Tensor) -> Tensor:
    """
    Computes the normalization constant for the Fisher Information Matrix (FIM).

    The function takes as input the integral over the trace of FIMs, the number
    of trainable parameters (i.e., the dimension of the FIM), and the volume of the
    parameter space.

    :param trace_integral: Integral over the trace of FIMs.
    :type trace_integral: Tensor
    :param num_parameters: Number of trainable parameters, or the dimension of the FIM.
    :type num_parameters: int
    :param volume: Volume of the parameter space.
    :type volume: Tensor

    :return: Normalization constant for the FIM.
    :rtype: Tensor

    :raises ValueError: If the trace integral is less than or equal to zero, the number of
        parameters is less than one, or the volume is negative.
    """

    if trace_integral <= 0.0:
        raise ValueError(f"trace_integral ({trace_integral}) must be positive")
    if num_parameters < 1:
        raise ValueError(f"num_parameters ({num_parameters}) has to be at least 1")
    if volume < 0.0:
        raise ValueError(f"volume ({volume}) has to be non-negative")

    norm_const = num_parameters * volume / trace_integral
    return norm_const


def mc_integrate_fim_trace(
    fims: List[Tensor],
    weights: Tensor,
) -> Tensor:
    """
    Performs a weighted Monte Carlo integration of the traces of the Fisher Information
    Matrices (FIMs), which are estimated at different parameter samples.

    The function takes as input a list of Fisher Information Matrices and weights for each
    parameter sample (or a single constant weight). It returns a scalar tensor of the
    estimated integral of the traces of the FIMs.

    :param fims: List of Fisher Information Matrices estimated at different parameter samples.
    :type fims: List[Tensor]
    :param weights: Weights for each parameter sample, or a single constant weight.
        The integral f(theta) is estimated as 1/N * sum_k f_k w_k.
    :type weights: Tensor

    :return: A scalar tensor of the estimated integral of the trace.
    :rtype: Tensor
    """

    num_samples = len(fims)
    if weights.numel() > 1:
        num_weights = len(weights)
        if num_samples != num_weights:
            raise ValueError(
                f"Number of samples ({num_samples}) "
                f"must be equal to number of weights ({num_weights})"
            )

    sum = torch.zeros(1, device=fims[0].device, dtype=fims[0].dtype)
    for i, fim in enumerate(fims):
        _check_fim(fim)
        trace = torch.trace(fim)

        if weights.numel() > 1:
            weight = weights[i]
        else:
            weight = weights

        sum += weight * trace

    sum /= num_samples

    return sum


def compute_fims(
    model: Model,
    features: Tensor,
    param_list: ParameterList,
) -> List[Tensor]:
    """
    Computes a list of empirical Fisher Information Matrices (FIMs) for a given parameter list.

    This function modifies the parameters of the model temporarily for each set of parameters
    in the parameter list and computes the corresponding empirical FIM.

    :param model: The model returning a discrete probability distribution.
    :type model: Model
    :param features: The input data (X) used to compute the FIMs.
    :type features: Tensor
    :param param_list: A list of lists of parameters. Each set of parameters is used to
        compute an empirical FIM.
    :type param_list: ParameterList

    :return: A list of Fisher information matrices.
    :rtype: List[Tensor]

    .. note::
        This function assumes that the model's parameters are differentiable.
    """

    original_params = [p.clone() for p in model.parameters() if p.requires_grad]

    fims = []
    for param in param_list:
        for model_param, sample_param in zip(model.parameters(), param):
            model_param.data = sample_param

        fim = empirical_fim(model, features)
        fims.append(fim)

    for model_param, original_param in zip(model.parameters(), original_params):
        model_param.data = original_param

    return fims


def empirical_fim(model: Model, features: Tensor) -> Tensor:
    """
    Computes the empirical Fisher information matrix.

    :param model: The model returning a discrete probability distribution.
    :type model: Model
    :param features: The training feature set X to estimate the FIM.
    :type features: Tensor

    :returns: The Fisher information matrix.
    :rtype: Tensor

    :raises ValueError: If invalid features format.

    .. note::
        This function assumes that the output of the model is a differentiable
        tensor of probabilities.
    """

    _check_features(features)

    probs = model(features)
    log_probs = torch.log(probs)
    num_parameters = sum(p.numel() for p in model.parameters() if p.requires_grad)
    num_samples = log_probs.shape[0]
    num_states = log_probs.shape[1]
    FIM = torch.zeros(
        (num_parameters, num_parameters), device=features.device, dtype=features.dtype
    )

    for sample in range(num_samples):
        for state in range(num_states):
            model.zero_grad()
            log_prob = log_probs[sample, state]
            log_prob.backward(retain_graph=True)
            grad_list = [
                p.grad.view(-1)
                for p in model.parameters()
                if p.requires_grad and p.grad is not None
            ]
            grad = torch.cat(grad_list)
            prod = torch.outer(grad, grad)
            FIM += prod * probs[sample, state]
    FIM /= num_samples

    return FIM


def _check_features(features: Tensor) -> None:
    lshapex = len(features.shape)
    if lshapex != 2:
        raise ValueError(f"features must be a 2-dim tensor (not {lshapex})")


def _check_fim(fim: Tensor) -> None:
    shape = fim.shape
    len_shape = len(shape)
    if len_shape != 2:
        raise ValueError(f"fim (shape={shape}) must be a squared matrix.")
    if shape[0] != shape[1]:
        raise ValueError(f"fim (shape={shape}) must be a squared matrix.")
